/**
 * @file acl.c ACL functions and definitions
 *
 * gaim
 *
 * Gaim is the legal property of its developers, whose names are too numerous
 * to list here.  Please refer to the COPYRIGHT file distributed with this
 * source distribution.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include "internal.h"

#include "account.h"
#include "debug.h"
#include "blist.h"
#include "util.h"
#include "xmlnode.h"
#include "privacy.h"

#include "session.h"
#include "httpconn.h"
#include "acl.h"
#include "subscribe.h"

void rvp_acl_cb(RvpHttpconn *hc, const gchar *receive)
{
  GaimConnection *gc = NULL;
  RvpSession *session = NULL;
  GaimBuddy *buddy = NULL;

  gchar *tmp = NULL;
  xmlnode *node = NULL;
  xmlnode *nodeace = NULL;
  xmlnode *nodebuddy = NULL;

  gchar *url = NULL;
  gchar *host = NULL;
  gchar *path = NULL;
  gchar *user = NULL;
  gchar *pwd = NULL;
  gchar *mail = NULL;
  int port = -1;

  GaimBlistNode *gnode = NULL;
  GaimBlistNode *cnode = NULL;
  GaimBlistNode *bnode = NULL;
  gboolean subscribe = TRUE;

  g_return_if_fail(hc != NULL);
  g_return_if_fail(receive != NULL);

  gc = hc->gc;
  g_return_if_fail(gc != NULL);

  session = gc->proto_data;
  g_return_if_fail(session != NULL);

  if (hc->inpa > 0)
    gaim_input_remove(hc->inpa);

  if (hc->fd > 0)
    close(hc->fd);

  if (strncmp(receive, "HTTP/1.1 200 Successful", 23) != 0)
    {
      gaim_debug(GAIM_DEBUG_ERROR, "rvp", "rvp_acl_cb failed\n");
      return;
    }

  gaim_debug(GAIM_DEBUG_INFO, "rvp", "rvp_acl_cb success\n");

  tmp = strstr(receive, "<?xml version=\"1.0\"?>");
  g_return_if_fail(tmp != NULL);

  node = xmlnode_from_str(tmp, strlen(tmp));
  g_return_if_fail(node != NULL);

  nodeace = xmlnode_get_child(node, "a:acl/a:ace");

  while (nodeace != NULL)
    {
      nodebuddy = xmlnode_get_child(nodeace,
                                    "a:principal/a:rvp-principal");
      if (nodebuddy == NULL)
        {
          nodeace = xmlnode_get_next_twin(nodeace);
          continue;
        }
      url = xmlnode_get_data(nodebuddy);
      if (url == NULL)
        {
          nodeace = xmlnode_get_next_twin(nodeace);
          continue;
        }

      gaim_url_parse(url, &host, &port, &path, &user, &pwd);
      tmp = strstr(path, "instmsg/aliases/");
      g_return_if_fail(tmp != NULL);
      tmp += strlen("instmsg/aliases/");

      mail = g_strdup_printf("%s@%s", tmp, host);

      g_free(host);
      g_free(path);
      g_free(user);
      g_free(pwd);

      for(gnode = gaim_get_blist()->root; gnode; gnode = gnode->next)
        {
          if(!GAIM_BLIST_NODE_IS_GROUP(gnode))
            continue;
          for(cnode = gnode->child; cnode; cnode = cnode->next)
            {
              if(!GAIM_BLIST_NODE_IS_CONTACT(cnode))
                continue;
              for(bnode = cnode->child; bnode; bnode = bnode->next)
                {
                  if(!GAIM_BLIST_NODE_IS_BUDDY(bnode))
                    continue;
                  if(((GaimBuddy*)bnode)->account == gc->account)
                    {
                      if (strcmp(gaim_buddy_get_name((GaimBuddy*)bnode),
                                 mail) == 0)
                        {
                          gaim_debug(GAIM_DEBUG_INFO, "rvp", "rvp_acl_cb buddy %s already in internal list.\n",
                                     mail);
                          subscribe = FALSE;
                        }
                    }
                }
            }
        }

      if (subscribe)
        {
          buddy = gaim_buddy_new(session->account, mail, NULL);
          g_return_if_fail(buddy != NULL);

          gaim_blist_add_buddy(buddy, NULL, NULL, NULL);
          rvp_subscribe(session, buddy);
        }

      if ((nodebuddy = xmlnode_get_child(nodeace,
                                         "a:grant/a:send-to")) != NULL)
        {
          gaim_debug(GAIM_DEBUG_INFO, "rvp", "rvp_acl_cb buddy %s grant\n",
                     mail);
          gaim_privacy_deny_remove(gc->account, mail, TRUE);
          gaim_privacy_permit_add(gc->account, mail, TRUE);
        }
      else if ((nodebuddy = xmlnode_get_child(nodeace,
                                              "a:deny/a:send-to")) != NULL)
        {
          gaim_debug(GAIM_DEBUG_INFO, "rvp", "rvp_acl_cb buddy %s deny\n",
                     mail);
          gc->account->perm_deny = GAIM_PRIVACY_DENY_USERS;
          gaim_privacy_permit_remove(gc->account, mail, TRUE);
          gaim_privacy_deny_add(gc->account, mail, TRUE);
        }
      else
        {
          gaim_debug(GAIM_DEBUG_ERROR, "rvp", "rvp_acl_cb permition unknown for %s\n",
                     mail);
        }

      nodeace = xmlnode_get_next_twin(nodeace);
    }

  xmlnode_free(node);

  return;
}

void
rvp_get_acl(RvpSession *session, GaimBuddy *buddy, int action)
{
  GaimConnection *gc = NULL;

  RvpHttpconn *hc = g_new0(RvpHttpconn, 1);
  gchar buf[80];

  RvpBuddy *rvpb = NULL;

  g_return_if_fail(session != NULL);
  g_return_if_fail(session->connected);

  gc = gaim_account_get_connection(session->account);
  g_return_if_fail(gc != NULL);

  gaim_debug(GAIM_DEBUG_INFO, "rvp", "rvp_get_acl\n");

  hc->gc = gc;
  hc->connection_step = 1;

  if (gc->inpa > 0)
    gaim_input_remove(gc->inpa);

  hc->method = "ACL";
  hc->uri = g_strdup_printf("/instmsg/aliases/%s", session->user);
  hc->callback = FALSE;
  hc->host = g_strdup(session->host);
  hc->option = (gpointer *)buddy;
  switch(action)
    {
    case RVP_DENY:
      g_return_if_fail(buddy != NULL);

      rvpb = buddy->proto_data;
      g_return_if_fail(rvpb != NULL);

      hc->body = g_strdup_printf(
                                 "<?xml version=\"1.0\"?>\n"
                                 "<a:rvpacl xmlns:d='DAV:' xmlns:r='http://schemas.microsoft.com/rvp/' "
                                 "xmlns:a='http://schemas.microsoft.com/rvp/acl/'>"
                                 "<a:acl>"
                                 "<a:inheritance>none</a:inheritance>"
                                 "<a:ace>"
                                 "<a:principal>"
                                 "<a:rvp-principal>http://%s/instmsg/aliases/%s</a:rvp-principal>"
                                 "<a:credentials>"
                                 "<a:assertion/>"
                                 "<a:digest/>"
                                 "<a:ntlm/>"
                                 "</a:credentials>"
                                 "</a:principal>"
                                 "<a:grant>"
                                 "</a:grant>"
                                 "<a:deny>"
                                 "<a:send-to/>"
                                 "<a:presence/>"
                                 "</a:deny>"
                                 "</a:ace>"
                                 "</a:acl>"
                                 "</a:rvpacl>",
                                 rvpb->host,
                                 rvpb->user);
      break;
    case RVP_PERMIT:
      g_return_if_fail(buddy != NULL);

      rvpb = buddy->proto_data;
      g_return_if_fail(rvpb != NULL);

      hc->body = g_strdup_printf(
                                 "<?xml version=\"1.0\"?>\n"
                                 "<a:rvpacl xmlns:d='DAV:' xmlns:r='http://schemas.microsoft.com/rvp/' "
                                 "xmlns:a='http://schemas.microsoft.com/rvp/acl/'>"
                                 "<a:acl>"
                                 "<a:inheritance>none</a:inheritance>"
                                 "<a:ace>"
                                 "<a:principal>"
                                 "<a:rvp-principal>http://%s/instmsg/aliases/%s</a:rvp-principal>"
                                 "<a:credentials>"
                                 "<a:assertion/>"
                                 "<a:digest/>"
                                 "<a:ntlm/>"
                                 "</a:credentials>"
                                 "</a:principal>"
                                 "<a:grant>"
                                 "<a:send-to/>"
                                 "<a:presence/>"
                                 "</a:grant>"
                                 "<a:deny>"
                                 "</a:deny>"
                                 "</a:ace>"
                                 "</a:acl>"
                                 "</a:rvpacl>",
                                 rvpb->host,
                                 rvpb->user);
      break;
    case RVP_NONE:
    default:
      hc->body = g_strdup("");
    }
  hc->header = g_strdup_printf(
                               "ACL %s HTTP/1.1\r\n"
                               "RVP-Notifications-Version: 0.2\r\n"
                               "Host: %s\r\n"
                               "Content-Length: %d\r\n"
                               "RVP-From-Principal: http://%s/instmsg/aliases/%s\r\n",
                               hc->uri,
                               hc->host,
                               strlen(hc->body),
                               session->host,
                               session->user);
  hc->analyze = rvp_acl_cb;

  if (gaim_proxy_connect(session->account, hc->host, RVP_PORT,
                         rvp_http_send, hc) != 0)
    {
      g_snprintf(buf, sizeof buf, _("Connect to %s failed"), hc->host);
      gaim_connection_error(gc, buf);
    }

  return;
}

void rvp_get_acl_buddies(GaimConnection *gc)
{
  GaimBlistNode *gnode = NULL;
  GaimBlistNode *cnode = NULL;
  GaimBlistNode *bnode = NULL;

  g_return_if_fail(gc != NULL);

  gaim_debug(GAIM_DEBUG_INFO, "rvp","rvp_get_buddies\n");

  for(gnode = gaim_get_blist()->root; gnode; gnode = gnode->next)
    {
      if(!GAIM_BLIST_NODE_IS_GROUP(gnode))
        continue;
      for(cnode = gnode->child; cnode; cnode = cnode->next)
        {
          if(!GAIM_BLIST_NODE_IS_CONTACT(cnode))
            continue;
          for(bnode = cnode->child; bnode; bnode = bnode->next)
            {
              if(!GAIM_BLIST_NODE_IS_BUDDY(bnode))
                continue;
              if(((GaimBuddy*)bnode)->account == gc->account)
                {
                  rvp_subscribe((RvpSession *)gc->proto_data, (GaimBuddy*) bnode);
                }
            }
        }
    }
  rvp_get_acl((RvpSession *)gc->proto_data, NULL, -1);
  return;
}

void rvp_set_acl_deny(RvpSession *session, GaimBuddy *buddy)
{
  g_return_if_fail(session != NULL);
  g_return_if_fail(buddy != NULL);

  rvp_get_acl(session, buddy, RVP_DENY);
  return;
}

void rvp_set_acl_permit(RvpSession *session, GaimBuddy *buddy)
{
  g_return_if_fail(session != NULL);
  g_return_if_fail(buddy != NULL);

  rvp_get_acl(session, buddy, RVP_PERMIT);
  return;
}
