/**
 * @file authentication.c RVP authentication functions
 *
 * gaim
 *
 * Copyright (C) 2005 Fabien Carrion <fabien.carrion@gmail.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
/*
 * NTLM algorithm implemented as described in NTLM Authentication Scheme for HTTP
 * from Ronald Tschalar / 17. June 2003 / ronald@innovation.ch.
 */
#include "internal.h"

#include "ntlm.h"
#include "util.h"
#include "cipher.h"

#include "session.h"
#include "httpconn.h"
#include "authentication.h"

gchar *
rvp_encode_basic(RvpSession *session, RvpHttpconn *hc)
{
  gchar *result = NULL;
  gchar *base64hash = NULL;

  g_return_val_if_fail(session != NULL, NULL);
  g_return_val_if_fail(hc != NULL, NULL);

  if (session->domain != NULL)
    result = g_strdup_printf("%s/%s:%s",
                             session->domain,
                             session->user,
                             session->password);
  else
    result = g_strdup_printf("%s:%s",
                             session->user,
                             session->password);

  base64hash = gaim_base64_encode((guchar *)result, strlen(result) * sizeof(gchar));
  g_free(result);

  return base64hash;
}

gchar *
rvp_encode_digest(RvpSession *session, RvpHttpconn *hc)
{
  GaimCipher *cipher = NULL;
  GaimCipherContext *ctx = NULL;

  gchar *domain = NULL;
  gchar *options = NULL;
  gchar *options2 = NULL;
  gchar *result = NULL;
  gchar hashedDigest[33];
  guchar ha1[33], ha2[33];
  gchar *nc = "00000001";
  gchar cnonce[33];
  gint i = 0, v = 0;

  gchar *realm = NULL;
  gchar *nonce = NULL;
  gchar *qop = NULL;

  g_return_val_if_fail(session != NULL, NULL);
  g_return_val_if_fail(hc != NULL, NULL);

  cipher = gaim_ciphers_find_cipher("md5");

  options = strstr(hc->auth_key, "realm") + 2;
  options2 = strchr(options, '\"');
  realm = g_strndup(options, options - options2);

  options = strstr(hc->auth_key, "nonce") + 2;
  options2 = strchr(options, '\"');
  nonce = g_strndup(options, options - options2);

  options = strstr(hc->auth_key, "qop") + 2;
  options2 = strchr(options, '\"');
  qop = g_strndup(options, options - options2);

  for (i=0; i<32; i++)
    {
      v = (rand() >> 2) % 0x0F;
      if (v > 9)
        cnonce[i] = 'a' + v - 10;
      else
        cnonce[i] = '0' + v;
    }
  cnonce[32]='\0';

  ctx = gaim_cipher_context_new(cipher, NULL);
  if (session->domain != NULL)
    {
      gaim_cipher_context_append(ctx, (const guchar *)session->domain, strlen(session->domain));
      gaim_cipher_context_append(ctx, (const guchar *)"/", 1);
    }
  gaim_cipher_context_append(ctx, (const guchar *)session->user, strlen(session->user));
  gaim_cipher_context_append(ctx, (const guchar *)":", 1);
  gaim_cipher_context_append(ctx, (const guchar *)realm, strlen(realm));
  gaim_cipher_context_append(ctx, (const guchar *)":", 1);
  gaim_cipher_context_append(ctx, (const guchar *)session->password, strlen(session->password));
  gaim_cipher_context_digest(ctx, sizeof(ha1), ha1, NULL);
    
  gaim_cipher_context_reset(ctx, NULL);
  gaim_cipher_context_append(ctx, (const guchar *)hc->method, strlen(hc->method));
  gaim_cipher_context_append(ctx, (const guchar *)":", 1);
  gaim_cipher_context_append(ctx, (const guchar *)hc->uri, strlen(hc->uri));
  gaim_cipher_context_digest(ctx, sizeof(ha2), ha2, NULL);

  gaim_cipher_context_reset(ctx, NULL);
  gaim_cipher_context_append(ctx, (const guchar *)ha1, 32);
  gaim_cipher_context_append(ctx, (const guchar *)":", 1);
  gaim_cipher_context_append(ctx, (const guchar *)nonce, strlen(nonce));
  if (qop != NULL)
    {
      gaim_cipher_context_append(ctx, (const guchar *)":", 1);
      gaim_cipher_context_append(ctx, (const guchar *)nc, strlen(nc));
      gaim_cipher_context_append(ctx, (const guchar *)":", 1);
      gaim_cipher_context_append(ctx, (const guchar *)cnonce, strlen(cnonce));
      gaim_cipher_context_append(ctx, (const guchar *)":", 1);
      gaim_cipher_context_append(ctx, (const guchar *)qop, strlen(qop));
    }
  gaim_cipher_context_append(ctx, (const guchar *)":", 1);
  gaim_cipher_context_append(ctx, (const guchar *)ha2, 32);
  gaim_cipher_context_digest(ctx, sizeof(hashedDigest), (guchar *)hashedDigest, NULL);

  gaim_cipher_context_destroy(ctx);

  if (session->domain != NULL)
    {
      domain = g_strdup_printf(
                               "username=\"%s/%s\", ",
                               session->domain,
                               session->user);
    }
  else
    {
      domain = g_strdup_printf(
                               "username=\"%s\", ",
                               session->user);
    }
  options = g_strdup_printf(
                            "realm=\"%s\", "
                            "algorithm=\"MD5\", "
                            "uri=\"%s\", "
                            "nonce=\"%s\", ",
                            realm,
                            hc->uri,
                            nonce);
  if (qop)
    {
      options2 = g_strdup_printf(
                                 "qop=\"%s\", "
                                 "nc=\"%s\", "
                                 "cnonce=\"%s\", ",
                                 qop,
                                 nc,
                                 cnonce);
    }
  result = g_strdup_printf(
                           "%s%s%s"
                           "response=\"%s\"",
                           domain,
                           options,
                           options2,
                           hashedDigest);
  g_free(realm);
  g_free(nonce);
  g_free(qop);

  g_free(domain);
  g_free(options);
  g_free(options2);

  return result;
}

gchar *
rvp_encode_ntlm_initialise(RvpSession *session, RvpHttpconn *hc)
{
  gchar *type1 = NULL;

  g_return_val_if_fail(session != NULL, NULL);
  g_return_val_if_fail(hc != NULL, NULL);

  type1 = gaim_ntlm_gen_type1("", session->domain);

  return type1;
}

gchar *
rvp_encode_ntlm_challenge(RvpSession *session, RvpHttpconn *hc)
{
  gchar *nonce = NULL;
  gchar *type3 = NULL;

  g_return_val_if_fail(session != NULL, NULL);
  g_return_val_if_fail(hc != NULL, NULL);

  nonce = gaim_ntlm_parse_type2(hc->auth_key + strlen("NTLM "));
  type3 = gaim_ntlm_gen_type3(session->user, session->password, "", session->domain, nonce);

  return type3;
}

gchar *
rvp_encode_password(RvpSession *session, RvpHttpconn *hc)
{
  gchar *tmp = NULL;
  gchar *ret = NULL;

  g_return_val_if_fail(session != NULL, NULL);

  if (hc->auth_key == NULL)
    {
      ret = g_strdup("");
    }
  else if (strstr(hc->auth_key, "NTLM ") != NULL)
    {
      tmp = rvp_encode_ntlm_challenge(session, hc);
      ret = g_strdup_printf("Authorization: NTLM %s\r\n", tmp);
    }
  else if (strstr(hc->auth_key, "NTLM") != NULL)
    {
      tmp = rvp_encode_ntlm_initialise(session, hc);
      ret = g_strdup_printf("Authorization: NTLM %s\r\n", tmp);
    }
  else if (strstr(hc->auth_key, "Digest ") != NULL)
    {
      tmp = rvp_encode_digest(session, hc);
      ret = g_strdup_printf("Authorization: Digest %s\r\n", tmp);
    }
  else if (strstr(hc->auth_key, "Basic") != NULL)
    {
      tmp = rvp_encode_basic(session, hc);
      ret = g_strdup_printf("Authorization: Basic %s\r\n", tmp);
    }
  else
    {
      ret = g_strdup("");
    }

  g_free(tmp);
  return ret;
}
