/**
 * @file file.c RVP file transfert functions
 *
 * gaim
 *
 * Copyright (C) 2005 Fabien Carrion <fabien.carrion@gmail.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include "internal.h"

#include "debug.h"
#include "ft.h"

#include "session.h"
#include "msg.h"
#include "file.h"

int
CookieCompare(gconstpointer xfer, gconstpointer tmp)
{
  gint cookie = GPOINTER_TO_INT(tmp);

  g_return_val_if_fail(xfer != NULL, -1);

  if (((RvpFile *)((GaimXfer *)xfer)->data)->cookie == cookie)
    return 0;

  return 1;
}


GaimXfer *
rvp_xfer_find(GaimBuddy *buddy, int cookie)
{
  GSList *xferl = NULL;
  RvpBuddy *rvpb = NULL;

  g_return_val_if_fail(buddy != NULL, NULL);

  rvpb = buddy->proto_data;
  g_return_val_if_fail(rvpb != NULL, NULL);

  gaim_debug(GAIM_DEBUG_INFO, "rvp", "rvp_xfer_find\n");

  xferl = g_slist_find_custom(rvpb->xfer, GINT_TO_POINTER(cookie),
                              (GCompareFunc) CookieCompare);

  if (xferl == NULL)
    return NULL;

  return xferl->data;
}

RvpFile *
rvp_file_new(void)
{
  RvpFile *file = g_new(RvpFile, 1);

  file->inpa1 = -1;
  file->listenfd1 = -1;
  file->inpa2 = -1;
  file->listenfd2 = -1;
  file->fd = -1;
  file->cookie = -1;
  file->authcookie = -1;
  file->ip = NULL;
  file->port1 = -1;
  file->port2 = -1;
  file->headersize = 0;
  file->canceled = FALSE;

  return file;
}


void
rvp_xfer_init(GaimXfer *xfer)
{
  GaimAccount *account = NULL;
  GaimBuddy *buddy = NULL;
  RvpBuddy *rvpb = NULL;
  const char *filename;

  g_return_if_fail(xfer != NULL);

  account = gaim_xfer_get_account(xfer);
  g_return_if_fail(account != NULL);

  buddy = gaim_find_buddy(account, xfer->who);
  g_return_if_fail(buddy != NULL);

  rvpb = buddy->proto_data;
  g_return_if_fail(rvpb != NULL);

  filename = gaim_xfer_get_local_filename(xfer);

  gaim_debug(GAIM_DEBUG_INFO, "rvp", "rvp_xfer_init_send %s\n",
             filename);

  rvpb->xfer = g_slist_prepend(rvpb->xfer, xfer);

  if (gaim_xfer_get_type(xfer) == GAIM_XFER_SEND)
    rvp_send_request_file(rvpb->session, buddy, xfer);
  if (gaim_xfer_get_type(xfer) == GAIM_XFER_RECEIVE)
    rvp_send_accept_file(rvpb->session, buddy, xfer);

  return;
}

void
rvp_xfer_start_send(GaimXfer *xfer)
{
  RvpFile *file = NULL;
  guchar buff[256];
  guchar tmp1[128];
  int tmp2 = -1;

  g_return_if_fail(xfer    != NULL);

  file = xfer->data;
  g_return_if_fail(file   != NULL);

  gaim_debug(GAIM_DEBUG_INFO, "rvp", "rvp_xfer_start_send %s\n",
             gaim_xfer_get_local_filename(xfer));

  memset(buff, '\0', 256);
  read(xfer->fd, buff, 256);
  if (strncmp(buff, "VER MSNFTP\r\n", 12))
    {
      gaim_debug(GAIM_DEBUG_ERROR, "rvp",
                 "rvp_xfer_start_send Client does not know MSNFTP\n");
      gaim_xfer_cancel_local(xfer);
      return;
    }
  write(xfer->fd, "VER MSNFTP\r\n", 12);

  memset(buff, '\0', 256);
  read(xfer->fd, buff, 256);
  sscanf(buff, "USR %s %d\r\n", tmp1, &tmp2);
  if (strcmp(tmp1, xfer->who))
    {
      gaim_debug(GAIM_DEBUG_ERROR, "rvp",
                 "rvp_xfer_start_send Bad receiver %s\n",
                 tmp1);
      gaim_xfer_cancel_local(xfer);
      return;
    }
  if (tmp2 != file->authcookie)
    {
      gaim_debug(GAIM_DEBUG_ERROR, "rvp",
                 "rvp_xfer_start_send Bad authcookie %s\n",
                 tmp2);
      gaim_xfer_cancel_local(xfer);
      return;
    }

  memset(buff, '\0', 256);
  g_snprintf(buff, 256, "FIL %d\r\n",
             gaim_xfer_get_size(xfer));
  write(xfer->fd, buff, strlen(buff));

  memset(buff, '\0', 256);
  read(xfer->fd, buff, 256);
  if (strncmp(buff, "TFR\r\n", 5))
    {
      gaim_debug(GAIM_DEBUG_ERROR, "rvp",
                 "rvp_xfer_start_send Client does not ask for file transfert\n");
      gaim_xfer_cancel_local(xfer);
      return;
    }
  return;
}

void
rvp_xfer_start_recv(GaimXfer *xfer)
{
  GaimAccount *account = NULL;
  GaimConnection *gc = NULL;
  RvpSession *session = NULL;
  RvpFile *file = NULL;
  guchar buff[256];
  int tmp1 = -1;

  g_return_if_fail(xfer    != NULL);

  account = gaim_xfer_get_account(xfer);
  g_return_if_fail(account != NULL);

  gc = gaim_account_get_connection(account);
  g_return_if_fail(gc != NULL);

  session = gc->proto_data;
  g_return_if_fail(session != NULL);

  file = xfer->data;
  g_return_if_fail(file    != NULL);

  gaim_debug(GAIM_DEBUG_INFO, "rvp", "rvp_xfer_start_recv %s\n",
             gaim_xfer_get_local_filename(xfer));

  write(xfer->fd, "VER MSNFTP\r\n", 12);

  memset(buff, '\0', sizeof(char) * 256);
  read(xfer->fd, buff, 256);
  if (strncmp(buff, "VER MSNFTP\r\n", 12))
    {
      gaim_debug(GAIM_DEBUG_ERROR, "rvp",
                 "rvp_xfer_start_recv Client does not know MSNFTP\n");
      gaim_xfer_cancel_local(xfer);
      return;
    }

  memset(buff, '\0', sizeof(char) * 256);
  g_snprintf(buff, sizeof(buff), "USR %s@%s %d\r\n",
             session->user,
             session->host,
             file->authcookie);
  write(xfer->fd, buff, strlen(buff));

  memset(buff, '\0', sizeof(char) * 256);
  read(xfer->fd, buff, 256);
  if (strncmp(buff, "FIL ", 4))
    {
      gaim_debug(GAIM_DEBUG_ERROR, "rvp",
                 "rvp_xfer_start_recv Client does not send the file size\n");
      gaim_xfer_cancel_local(xfer);
      return;
    }
  sscanf(buff, "FIL %d\r\n", &tmp1);
  gaim_xfer_set_size(xfer, tmp1);

  write(xfer->fd, "TFR\r\n", 5);

  return;
}

void
rvp_xfer_end(GaimXfer *xfer)
{
  GaimBuddy *buddy = NULL;
  RvpBuddy *rvpb = NULL;
  RvpFile *file = NULL;

  g_return_if_fail(xfer != NULL);

  buddy = gaim_find_buddy(gaim_xfer_get_account(xfer),
                          xfer->who);
  g_return_if_fail(buddy != NULL);

  rvpb = buddy->proto_data;
  g_return_if_fail(rvpb != NULL);

  gaim_debug(GAIM_DEBUG_INFO, "rvp", "rvp_xfer_end %s\n",
             gaim_xfer_get_local_filename(xfer));

  rvpb->xfer = g_slist_remove(rvpb->xfer, xfer);

  file = xfer->data;

  if (file != NULL)
    {
      if (file->fd > 0)
        close(file->fd);

      if (file->inpa1 > 0)
        gaim_input_remove(file->inpa1);
      if (file->listenfd1 > 0)
        close(file->listenfd1);

      if (file->inpa2 > 0)
        gaim_input_remove(file->inpa2);
      if (file->listenfd2 > 0)
        close(file->listenfd2);

      xfer->data = NULL;
      g_free(file->ip);
      g_free(file);
    }

  return;
}

ssize_t
rvp_xfer_read(guchar **buffer, GaimXfer *xfer)
{
  RvpFile *file = NULL;
  ssize_t r = 0;
  ssize_t s = 0;

  g_return_val_if_fail(xfer    != NULL, 0);
  g_return_val_if_fail(buffer  != NULL, 0);

  file = xfer->data;
  g_return_val_if_fail(file    != NULL, 0);

  gaim_debug(GAIM_DEBUG_INFO, "rvp", "rvp_xfer_read %s\n",
             gaim_xfer_get_local_filename(xfer));
  printf("rvp rvp_xfer_read\n");

  if (file->headersize == 0)
    {
      guchar tmp[3];
      read(xfer->fd, tmp, 3);
      s = tmp[2] * 256 + tmp[1];
      file->headersize = s;

      printf("packet type %d\n", tmp[0]);
      switch(tmp[0])
        {
        case 0:
          printf("send packet\n");
          break;
        case 1:
          printf("end packet\n");
          break;
        default:
          printf("unknow packet\n");
          break;
        }

      printf("packet size %d = %d * 256 + %d\n", s, tmp[2], tmp[1]);
    }
  else
    {
      s = file->headersize;
    }

  *buffer = g_malloc0(s);
  r = read(xfer->fd, *buffer, s);
  file->headersize -= r;

  printf("Stat :\n");
  printf("gaim_xfer_get_size = %d\n", gaim_xfer_get_size(xfer));
  printf("gaim_xfer_get_bytes_sent = %d\n", gaim_xfer_get_bytes_sent(xfer));
  printf("asked for = %d\n", s);
  printf("received this time = %d\n", r);
  if ((gaim_xfer_get_size(xfer) > 0) &&
      ((gaim_xfer_get_bytes_sent(xfer)+r) >= gaim_xfer_get_size(xfer)))
    {
      /*read(xfer->fd, tmp, 3);
      if (strncmp(tmp, "\x01\x00\x00", 3))
        {
          gaim_debug(GAIM_DEBUG_ERROR, "rvp",
                     "rvp_xfer_read Client does not end the session\n");
          //return -1;
          }*/

      write(xfer->fd, "BYE 16777989\r\n", 14);

      gaim_xfer_set_completed(xfer, TRUE);
    }

  return r;
}

ssize_t
rvp_xfer_write(const guchar *buffer, size_t size, GaimXfer *xfer)
{
  guchar *buff;
  ssize_t r, s;

  g_return_val_if_fail(xfer   != NULL, 0);
  g_return_val_if_fail(buffer != NULL, 0);
  g_return_val_if_fail(size   != 0,    0);

  gaim_debug(GAIM_DEBUG_INFO, "rvp", "rvp_xfer_write %s\n",
             gaim_xfer_get_local_filename(xfer));

  s = MIN(gaim_xfer_get_bytes_remaining(xfer), size);

  buff = g_malloc0(s + 3);

  buff[0] = 0;
  buff[1] = s % 256;
  buff[2] = s / 256;

  memcpy(&buff[3], buffer, s);
  r = write(xfer->fd, buff, s + 3);
  r -= 3;

  if ((gaim_xfer_get_bytes_sent(xfer)+r) >= gaim_xfer_get_size(xfer))
    {
      write(xfer->fd, "\x01\x00\x00", 3);

      memset(buff, '\0', s+3);
      read(xfer->fd, buff, s+3);
      if (strncmp(buff, "BYE ", 4))
        if (strncmp(buff, "CCL\r\n", 5))
          {
            gaim_debug(GAIM_DEBUG_ERROR, "rvp",
                       "rvp_xfer_write Client does not end the session\n");
            return -1;
          }

      gaim_xfer_set_completed(xfer, TRUE);
    }

  g_free(buff);
  return r;
}


void
rvp_xfer_cancel(GaimXfer *xfer)
{
  gaim_debug(GAIM_DEBUG_INFO, "rvp", "rvp_xfer_cancel %s\n",
             gaim_xfer_get_local_filename(xfer));

  rvp_xfer_end(xfer);

  return;
}

void
rvp_xfer_cancel_notify(GaimXfer *xfer)
{
  GaimBuddy *buddy = NULL;
  RvpBuddy *rvpb = NULL;
  RvpFile *file = NULL;

  g_return_if_fail(xfer != NULL);

  buddy = gaim_find_buddy(gaim_xfer_get_account(xfer),
                          xfer->who);
  g_return_if_fail(buddy != NULL);

  rvpb = buddy->proto_data;
  g_return_if_fail(rvpb != NULL);

  file = xfer->data;
  g_return_if_fail(file != NULL);

  gaim_debug(GAIM_DEBUG_INFO, "rvp", "rvp_xfer_cancel_notify %s\n",
             gaim_xfer_get_local_filename(xfer));

  if (file->canceled == TRUE)
    return;

  file->canceled = TRUE;
  rvp_send_cancel_timeout_file(rvpb->session, buddy, xfer);

  rvp_xfer_end(xfer);

  return;
}
