/**
 * @file httpconn.h RVP http connection functions
 *
 * gaim
 *
 * Copyright (C) 2005 Fabien Carrion <fabien.carrion@gmail.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include "internal.h"

#include "account.h"
#include "proxy.h"
#include "debug.h"
#include "network.h"

#include "session.h"
#include "httpconn.h"
#include "authentication.h"


gchar *
getHeader(const gchar *receive, const gchar *header)
{
  gchar *tmp1 = NULL;
  gchar *tmp2 = NULL;

  g_return_val_if_fail(receive != NULL, NULL);
  g_return_val_if_fail(header != NULL, NULL);

  if ((tmp1 = strstr(receive, header)) == NULL)
    return NULL;

  tmp1 += strlen(header);
  tmp2 = strstr(tmp1, "\r\n");
  return g_strndup(tmp1, tmp2 - tmp1);
}


void
hc_clean(RvpHttpconn *hc)
{
  g_free(hc->uri);
  g_free(hc->auth_key);

  g_free(hc->host);
  g_free(hc->header);
  g_free(hc->body);
  g_free(hc);

  return;
}


void
rvp_http_send(gpointer data, gint source, GaimInputCondition condition)
{
  RvpHttpconn *hc = NULL;
  GaimConnection *gc = NULL;
  RvpSession *session = NULL;

  gchar *request = NULL;
  gchar *password = NULL;
  gchar *callback = NULL;
  gchar buf[80];

  g_return_if_fail(data != NULL);

  hc = data;
  g_return_if_fail(hc != NULL);

  gc = hc->gc;
  g_return_if_fail(gc != NULL);

  session = gc->proto_data;
  g_return_if_fail(session != NULL);

  if (source <= 0)
    {
      gaim_debug(GAIM_DEBUG_ERROR, "rvp",
                 "rvp_http_send connect failed on %s\n", hc->host);
      return;
    }

  hc->fd = source;
  fcntl(hc->fd, F_SETFL, O_NONBLOCK);

  if (gc->inpa > 0)
    gaim_input_remove(gc->inpa);

  if (hc->inpa > 0)
    gaim_input_remove(hc->inpa);

  if (hc->callback == TRUE)
    {
      if (session->ip == NULL)
        session->ip = gaim_network_get_my_ip(hc->fd);
      callback = g_strdup_printf(
                                 "Call-Back: http://%s:%d\r\n",
                                 session->ip,
                                 gaim_network_get_port_from_fd(session->fd));
    }
  else
    {
      callback = g_strdup(
                          "");
    }

  if (hc->connection_step == 1)
    hc->auth_key = NULL;

  password = rvp_encode_password(session, hc);
  g_return_if_fail(password != NULL);

  hc->connection_step++;
  if (session->connected == FALSE)
    {
      g_snprintf(buf, sizeof(buf), _("Signon: %s"), session->user);
      gaim_connection_update_progress(gc, buf, hc->connection_step, RVP_CONNECT_STEPS);
    }

  request = g_strdup_printf(
			    "%s"
                            "%s"
			    "%s\n"
                            "%s",
			    hc->header,
                            callback,
			    password,
                            hc->body);
  g_free(callback);
  g_free(password);

  gaim_debug(GAIM_DEBUG_INFO, "rvp", "Send :\n%s\n", request);
  write(hc->fd, request, strlen(request));
  g_free(request);

  hc->inpa = gaim_input_add(hc->fd, GAIM_INPUT_READ, rvp_http_receive, hc);

  return;
}

void
rvp_http_receive(gpointer data, gint source, GaimInputCondition condition)
{
  GaimConnection *gc = NULL;
  RvpHttpconn *hc = NULL;
  RvpSession *session = NULL;

  gchar *buffer = NULL;
  gchar buf[80];
  gchar cdata;

  gint len = 0;
  gint buffer_len = 4096;

  g_return_if_fail(data != NULL);

  hc = data;
  g_return_if_fail(hc != NULL);

  gc = hc->gc;
  g_return_if_fail(gc != NULL);

  session = gc->proto_data;
  g_return_if_fail(session != NULL);

  if (hc->inpa > 0)
    gaim_input_remove(hc->inpa);

  if (source <= 0)
    {
      gaim_debug(GAIM_DEBUG_ERROR, "rvp",
                 "rvp_http_receive connect failed on %s\n", hc->host);
      return;
    }
  hc->fd = source;

  buffer = g_new0(gchar, buffer_len);
  g_return_if_fail(buffer != NULL);

  /* there's data waiting to be read, so read it. */
  /* Read in data, one byte at a time */
  while (read(hc->fd, &cdata, 1) > 0 || errno == EWOULDBLOCK)
    {
      if (errno == EWOULDBLOCK)
        {
          gchar *tmp = NULL;
          int contentlength = 0;

          errno = 0;

          if ((tmp = getHeader(buffer, "Content-Length: ")) == NULL)
            break;

          contentlength = atoi(tmp);
          g_free(tmp);
          if ((tmp = strstr(buffer, "\r\n\r\n")) == NULL)
            break;
          tmp += strlen("\r\n\r\n");

          if (contentlength <= strlen(tmp))
            break;
        }
      len++;

      /* If we've filled up our buffer then make it bigger */
      if (len == buffer_len)
        {
          buffer_len += (buffer_len) / 2;
          buffer = g_realloc(buffer, buffer_len);
        }

      buffer[len - 1] = cdata;
      buffer[len] = '\0';
    }

  gaim_debug(GAIM_DEBUG_INFO, "rvp", "Received :\n%s\n", buffer);

  if (strstr(buffer, "Connection: close\r\n") != NULL)
    close(hc->fd);

  if (strncmp(buffer, "HTTP/1.1 401 Access Denied\r\n", 28) == 0)
    {
      gchar *auth_key = NULL;
      gchar *tmp = NULL;

      if ((tmp = strstr(buffer, "WWW-Authenticate: Negotiate\r\n")) != NULL)
	tmp += strlen("WWW-Authenticate: Negotiate\r\n");
      else
	tmp = buffer;
      if ((auth_key = strstr(tmp, "WWW-Authenticate: ")) == NULL)
	{
          gaim_debug(GAIM_DEBUG_ERROR, "rvp",
                     "rvp_http_receive Authentication Algorithm unknown\n");
	  return;
	}
      auth_key += strlen("WWW-Authenticate: ");

      tmp = strstr(auth_key, "\r\n");

      g_free(hc->auth_key);
      hc->auth_key = g_strndup(auth_key, tmp - auth_key);

      switch (hc->connection_step)
	{
	case 2:
	  hc->connection_step++;

	  if (session->connected == FALSE)
	    {
	      g_snprintf(buf, sizeof buf, _("Looking up %s"), session->host);
	      gaim_connection_update_progress(gc, buf, hc->connection_step, RVP_CONNECT_STEPS);
	    }

	  if (gaim_proxy_connect(session->account, hc->host, RVP_PORT,
				 rvp_http_send, hc) != 0)
	    {
              gaim_debug(GAIM_DEBUG_ERROR, "rvp",
                         "rvp_http_receive connect failed on %s\n", hc->host);
	    }
	  break;
	case 4:
	  hc->inpa = gaim_input_add(hc->fd, GAIM_INPUT_WRITE, rvp_http_send, hc);
	  break;
	default:
          hc->analyze(hc, buffer);
          hc_clean(hc);
        }
    }
  else
    {
      hc->analyze(hc, buffer);
      hc_clean(hc);
    }

  g_free(buffer);

  return;
}
