/**
 * @file incomconn.c RVP incomming connection functions
 *
 * gaim
 *
 * Copyright (C) 2005 Fabien Carrion <fabien.carrion@gmail.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include "internal.h"

#include "account.h"
#include "network.h"
#include "debug.h"
#include "xmlnode.h"
#include "util.h"
#include "ft.h"

#include "session.h"
#include "incomconn.h"
#include "httpconn.h"
#include "state.h"
#include "file.h"
#include "msg.h"

void
parse_message(RvpHttpconn *hc, GaimBuddy *buddy, gchar *message)
{
  RvpBuddy *rvpb = NULL;
  gchar *sessionid = NULL;
  gchar *action = NULL;
  gchar *tmp1 = NULL;
  gchar *tmp = NULL;

  g_return_if_fail(hc != NULL);
  g_return_if_fail(buddy != NULL);
  g_return_if_fail(message != NULL);

  rvpb = buddy->proto_data;
  g_return_if_fail(rvpb != NULL);

  tmp = strstr(message, "<![CDATA[MIME-Version: 1.0");
  g_return_if_fail(tmp != NULL);

  sessionid = getHeader(tmp, "Session-Id: ");

  if (sessionid != NULL)
    {
      if (rvpb->sessionid != NULL)
        g_free(rvpb->sessionid);
      rvpb->sessionid = sessionid;
    }

  gaim_debug(GAIM_DEBUG_INFO, "rvp", "parse_message\n");

  if ((action = getHeader(tmp, "TypingUser: ")) != NULL)
    {
      serv_got_typing(hc->gc, gaim_buddy_get_name(buddy), 5, GAIM_TYPING);
      g_free(action);
      return;
    }

  if (strstr(tmp, "Application-GUID: " RVP_GUID "\r\n") != NULL)
    {
      GaimXfer *xfer = NULL;
      RvpFile *file = rvp_file_new();

      xfer = gaim_xfer_new(hc->gc->account, GAIM_XFER_RECEIVE, gaim_buddy_get_name(buddy));
      g_return_if_fail(xfer != NULL);

      tmp1 = getHeader(tmp, "Invitation-Cookie: ");
      if (tmp1 == NULL)
        {
          gaim_debug(GAIM_DEBUG_ERROR, "rvp", "parse_message Invitation-Cookie unknown\n");
          return;
        }

      file->cookie = atoi(tmp1);
      g_free(tmp1);
      xfer->data = file;
      g_return_if_fail(file != NULL);

      tmp1 = getHeader(tmp, "Application-File: ");
      if (tmp1 == NULL)
        {
          gaim_debug(GAIM_DEBUG_ERROR, "rvp", "parse_message Application-File unknown\n");
          return;
        }

      gaim_xfer_set_filename(xfer, tmp1);

      tmp1 = getHeader(tmp, "Application-FileSize: ");
      if (tmp1 == NULL)
        {
          gaim_debug(GAIM_DEBUG_ERROR, "rvp", "parse_message Application-FileSize unknown\n");
          return;
        }

      gaim_xfer_set_size(xfer, atoi(tmp1));
      g_free(tmp1);

      /* Setup our I/O op functions */
      gaim_xfer_set_init_fnc(xfer, rvp_xfer_init);
      gaim_xfer_set_start_fnc(xfer, rvp_xfer_start_recv);
      gaim_xfer_set_end_fnc(xfer, rvp_xfer_end);
      gaim_xfer_set_read_fnc(xfer, rvp_xfer_read);
      gaim_xfer_set_request_denied_fnc(xfer, rvp_xfer_cancel);
      gaim_xfer_set_cancel_recv_fnc(xfer, rvp_xfer_cancel_notify);

      gaim_xfer_request(xfer);

      return;
    }

  if (strstr(tmp, "Invitation-Command: CANCEL\r\n") != NULL)
    {
      GaimXfer *xfer = NULL;
      int cookie = -1;

      tmp1 = getHeader(tmp, "Invitation-Cookie: ");
      if (tmp1 == NULL)
        {
          gaim_debug(GAIM_DEBUG_ERROR, "rvp", "parse_message Invitation-Cookie unknown\n");
          return;
        }

      cookie = atoi(tmp1);
      g_free(tmp1);
      if ((xfer = rvp_xfer_find(buddy, cookie)) == NULL)
        {
          gaim_debug(GAIM_DEBUG_ERROR, "rvp", "parse_message Invitation-Cookie unknown\n");
          return;
        }
      gaim_xfer_cancel_remote(xfer);

      return;
    }

  if (strstr(tmp, "Invitation-Command: ACCEPT\r\n") != NULL)
    {
      GaimXfer *xfer = NULL;
      RvpFile *file = NULL;
      int cookie = -1;

      tmp1 = getHeader(tmp, "Invitation-Cookie: ");
      if (tmp1 == NULL)
        {
          gaim_debug(GAIM_DEBUG_ERROR, "rvp", "parse_message Invitation-Cookie unknown\n");
          return;
        }
      cookie = atoi(tmp1);
      g_free(tmp1);

      if ((xfer = rvp_xfer_find(buddy, cookie)) == NULL)
        {
          gaim_debug(GAIM_DEBUG_ERROR, "rvp", "parse_message Invitation-Cookie unknown\n");
          return;
        }

      file = xfer->data;
      g_return_if_fail(file != NULL);

      tmp1 = getHeader(tmp, "AuthCookie: ");
      if (tmp1 == NULL)
        {
          file->listenfd1 = gaim_network_listen(RVP_PORT1);
          if (file->listenfd1 == -1)
            gaim_debug(GAIM_DEBUG_ERROR, "rvp", "parse_message Could not create listen socket");
          else
            file->inpa1 = gaim_input_add(file->listenfd1, GAIM_INPUT_READ, rvp_newfile_cb, xfer);

          file->listenfd2 = gaim_network_listen(RVP_PORT2);
          if (file->listenfd2 == -1)
            {
              gaim_debug(GAIM_DEBUG_ERROR, "rvp", "parse_message Could not create listen socket");
              gaim_xfer_cancel_local(xfer);
              return;
            }
          file->inpa2 = gaim_input_add(file->listenfd2, GAIM_INPUT_READ, rvp_newfile_cb, xfer);

          rvp_send_accept_info_file(rvpb->session, buddy, xfer);

          return;
        }
      file->authcookie = atoi(tmp1);
      g_free(tmp1);

      tmp1 = getHeader(tmp, "IP-Address: ");
      if (tmp1 == NULL)
        {
          gaim_debug(GAIM_DEBUG_ERROR, "rvp", "parse_message IP-Address unknown\n");
          return;
        }
      file->ip = tmp1;

      tmp1 = getHeader(tmp, "Port: ");
      if (tmp1 != NULL)
        {
          file->port1 = atoi(tmp1);
          g_free(tmp1);
        }

      tmp1 = getHeader(tmp, "PortX: ");
      if (tmp1 != NULL)
        {
          file->port2 = atoi(tmp1);
          g_free(tmp1);
        }

      gaim_xfer_start(xfer, -1, file->ip, file->port1);
      return;
    }

  if (strstr(tmp, "Content-Type: text/plain") != NULL)
    {
      tmp = strstr(tmp, "\r\n\r\n");
      tmp += strlen("\r\n\r\n");
      tmp1 = strstr(tmp, "]]>");
      tmp = g_strndup(tmp, tmp1 - tmp);

      serv_got_im(hc->gc, gaim_buddy_get_name(buddy),
                  gaim_utf8_salvage(tmp), 0, time(NULL));
      g_free(tmp);
      return;
    }

  return;
}


void
parse_receive(RvpHttpconn *hc, const gchar *receive)
{
  GaimConnection *gc = NULL;
  RvpSession *session = NULL;
  xmlnode *node = NULL;
  xmlnode *node1 = NULL;
  xmlnode *node2 = NULL;

  gchar *tmp = NULL;

  gchar *url = NULL;
  gchar *host = NULL;
  gchar *path = NULL;
  gchar *user = NULL;
  gchar *pwd = NULL;
  gchar *mail = NULL;

  GaimBuddy *buddy = NULL;
  int i = -1;

  g_return_if_fail(hc != NULL);
  g_return_if_fail(receive != NULL);

  gc = hc->gc;
  g_return_if_fail(gc != NULL);

  session = gc->proto_data;
  g_return_if_fail(session != NULL);

  gaim_debug(GAIM_DEBUG_INFO, "rvp", "parse_receive\n");

  tmp = strstr(receive, "<?xml version=\"1.0\"?>");
  g_return_if_fail(tmp != NULL);

  node = xmlnode_from_str(tmp, strlen(tmp));
  g_return_if_fail(node != NULL);

  if (strncmp(receive, "NOTIFY", 6) == 0)
    {
      if ((node1 = xmlnode_get_child(node,
                                     "r:message")) != NULL)
        {
          node2 = xmlnode_get_child(node1,
                                    "r:notification-from/r:contact/d:href");
          url = xmlnode_get_data(node2);
          g_return_if_fail(url != NULL);
          gaim_url_parse(url, &host, &i, &path, &user, &pwd);
          g_free(user);
          g_free(pwd);

          user = strstr(path, "instmsg/aliases/");
          user += strlen("instmsg/aliases/");

          mail = g_strdup_printf("%s@%s", user, host);
          g_free(host);

          buddy = gaim_find_buddy(session->account, mail);
          g_return_if_fail(buddy != NULL);

          node2 = xmlnode_get_child(node1,
                                    "r:notification-to/r:contact/d:href");
          g_return_if_fail(node2 != NULL);

          while (node2 != NULL)
            {
              GaimBuddy *b = NULL;

              g_free(path);
              g_free(mail);

              url = xmlnode_get_data(node2);
              g_return_if_fail(url != NULL);
              gaim_url_parse(url, &host, &i, &path, &user, &pwd);
              g_free(user);
              g_free(pwd);

              user = strstr(path, "instmsg/aliases/");
              user += strlen("instmsg/aliases/");

              mail = g_strdup_printf("%s@%s", user, host);
              g_free(host);

              if ((b = gaim_find_buddy(session->account, mail)) != NULL)
                {
                  gaim_debug(GAIM_DEBUG_INFO, "rvp", "parse_received chat with %s\n", mail);
                }

              node2 = xmlnode_get_next_twin(node2);
            }

          parse_message(hc, buddy, tmp);
        }
      else if ((node1 =
                xmlnode_get_child(node,
                                  "r:propnotification/r:notification-from/r:contact/d:href")) != NULL)
        {
          RvpAwayType state = RVP_AWAY;
          gint i = 0;

          url = xmlnode_get_data(node1);
          g_return_if_fail(url != NULL);
          gaim_url_parse(url, &host, &i, &path, &user, &pwd);
          g_free(user);
          g_free(pwd);

          user = strstr(path, "instmsg/aliases/");
          user += strlen("instmsg/aliases/");

          mail = g_strdup_printf("%s@%s", user, host);
          g_free(host);

          buddy = gaim_find_buddy(session->account, mail);
          g_return_if_fail(buddy != NULL);

          node1 = xmlnode_get_child(node,
                                    "r:message/r:propertyupdate/d:set/d:prop/r:state");
          g_return_if_fail(node1 != NULL);
          state = rvp_state_get_number(xmlnode_to_str(node1->child, &i));
          gaim_prpl_got_user_status(session->account, gaim_buddy_get_name(buddy),
                                    rvp_away_get_text(state),
                                    NULL);
        }
      g_free(path);
      g_free(mail);

    }

  xmlnode_free(node);

  return;
}

void
rvp_input_cb(RvpHttpconn *hc, const gchar *receive)
{
  GaimConnection *gc = NULL;
  RvpSession *session = NULL;
  gchar *request = NULL;

  g_return_if_fail(hc != NULL);
  g_return_if_fail(receive != NULL);

  gc = hc->gc;
  g_return_if_fail(gc != NULL);

  session = gc->proto_data;
  g_return_if_fail(session != NULL);

  request = g_strdup_printf(
			    "HTTP/1.1 200 Successful\r\n"
			    "Content-Type: text/html\r\n"
			    "RVP-Notifications-Version: 0.2\r\n"
			    "\r\n"
			    "<HTML>"
			    "<HEAD><TITLE>Successful</TITLE></HEAD>"
			    "<BODY>"
			    "<H2>Success 200 (Successful)</H2>"
			    "<HR>NOTIFY on node http://%s/instmsg/aliases/%s succeeded<HR>"
			    "</BODY>"
			    "</HTML>\r\n",
			    session->host,
			    session->user);

  gaim_debug(GAIM_DEBUG_INFO, "rvp", "rvp_input_cb:\n");

  parse_receive(hc, receive);

#if 0
  gaim_debug(GAIM_DEBUG_INFO, "rvp", "Send :\n%s\n", request);
#endif
  write(hc->fd, request, strlen(request));
  g_free(request);
  close(hc->fd);

  return;
}

void
rvp_newconn_cb(gpointer data, gint source, GaimInputCondition cond)
{
  GaimConnection *gc = NULL;
  RvpSession *session = NULL;

  int newfd = accept(source, NULL, NULL);

  RvpHttpconn *hc = g_new0(RvpHttpconn, 1);
  g_return_if_fail(hc != NULL);

  g_return_if_fail(data != NULL);
  gc = data;

  session = gc->proto_data;
  g_return_if_fail(session != NULL);

  gaim_debug(GAIM_DEBUG_INFO, "rvp", "rvp_newconn_cb\n");

  hc->gc = gc;
  hc->connection_step = 2;

  hc->fd = newfd;
  fcntl(hc->fd, F_SETFL, O_NONBLOCK);
  hc->analyze = rvp_input_cb;

  hc->inpa = gaim_input_add(newfd, GAIM_INPUT_READ, rvp_http_receive, hc);
}

void
rvp_newfile_cb(gpointer data, gint source, GaimInputCondition cond)
{
  GaimXfer *xfer = NULL;
  GaimBuddy *buddy = NULL;
  RvpBuddy *rvpb = NULL;
  RvpFile *file = NULL;

  int newfd = accept(source, NULL, NULL);

  xfer = data;
  buddy = gaim_find_buddy(gaim_xfer_get_account(xfer),
                          xfer->who);
  g_return_if_fail(buddy != NULL);

  rvpb = buddy->proto_data;
  g_return_if_fail(rvpb != NULL);

  file = xfer->data;
  g_return_if_fail(file != NULL);

  gaim_debug(GAIM_DEBUG_INFO, "rvp", "rvp_newfile_cb\n");

  file->fd = newfd;

  gaim_xfer_start(data, file->fd, NULL, -1);

  return;
}
