/**
 * @file msg.c RVP send NOTIFY message to buddy functions
 *
 * gaim
 *
 * Copyright (C) 2005 Fabien Carrion <fabien.carrion@gmail.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include "internal.h"

#include "account.h"
#include "debug.h"
#include "cipher.h"
#include "notify.h"
#include "util.h"
#include "ft.h"
#include "network.h"

#include "session.h"
#include "httpconn.h"
#include "file.h"
#include "msg.h"

void
rvp_notify_cb(RvpHttpconn *hc, const gchar *receive)
{
  GaimBuddy *b = NULL;

  g_return_if_fail(hc != NULL);
  g_return_if_fail(receive != NULL);

  b = hc->option;
  g_return_if_fail(b != NULL);

  if (hc->inpa > 0)
    gaim_input_remove(hc->inpa);

  if (hc->fd > 0)
    close(hc->fd);

  if ((strncmp(receive, "HTTP/1.1 412 Precondition Failed\r\n", 34) == 0) ||
      (strncmp(receive, "HTTP/1.1 500 Internal Server Error\r\n", 36) == 0))
    {
      gchar *buf = g_strdup_printf(_("buddy %s not reached."),
                                   gaim_buddy_get_name(b));
      gaim_notify_error(hc->gc, NULL, _("Unable To Send message"), buf);
      return;
    }

  if (strncmp(receive, "HTTP/1.1 200 Successful\r\n", 25) == 0)
    {
      gaim_debug(GAIM_DEBUG_INFO, "rvp", "rvp_notify_cb success %s\n",
                 gaim_buddy_get_name(b));

    }

  return;
}

void
rvp_notify(RvpSession *session, GaimBuddy *buddy, gchar *message,
           int rvphopcount, gchar *rvpacktype)
{
  GaimConnection *gc = NULL;

  RvpHttpconn *hc = g_new0(RvpHttpconn, 1);
  gchar buf[80];

  RvpBuddy *rvpb = NULL;

  gchar *urlprincipal = NULL;
  gchar *action = NULL;

  g_return_if_fail(session != NULL);
  g_return_if_fail(session->connected);
  g_return_if_fail(buddy != NULL);

  gc = gaim_account_get_connection(session->account);
  g_return_if_fail(gc != NULL);

  gaim_debug(GAIM_DEBUG_INFO, "rvp", "rvp_notify %s\n",
             gaim_buddy_get_name(buddy));

  hc->gc = gc;
  hc->connection_step = 1;

  if (gc->inpa > 0)
    gaim_input_remove(gc->inpa);

  rvpb = buddy->proto_data;
  g_return_if_fail(rvpb != NULL);

  switch (rvphopcount)
    {
    case 1:
      action = "message";
      hc->uri = g_strdup_printf(
				"/instmsg/aliases/%s",
				rvpb->user);
      urlprincipal = g_strdup_printf("http://%s/instmsg/aliases/%s",
                                     session->host,
                                     session->user);
      break;
    case 2:
      action = "propnotification";
      hc->uri = g_strdup("/");
      urlprincipal = g_strdup(session->host);
      break;
    case 3:
      gaim_debug(GAIM_DEBUG_ERROR, "rvp", "rvp_notify rvphopcount not implemented\n");
      return;
    default:
      gaim_debug(GAIM_DEBUG_ERROR, "rvp", "rvp_notify rvphopcount unknown\n");
      return;
    }

  hc->method = "NOTIFY";
  hc->callback = FALSE;
  hc->host = g_strdup(rvpb->host);
  hc->option = (gpointer *)buddy;
  hc->body = g_strdup_printf(
                             "<?xml version=\"1.0\"?>\n"
                             "<r:notification xmlns:d='DAV:' "
                             "xmlns:r='http://schemas.microsoft.com/rvp/' "
                             "xmlns:a='http://schemas.microsoft.com/rvp/acl/'>"
                             "<r:%s>"
                             "<r:notification-from>"
                             "<r:contact>"
                             "<d:href>"
                             "http://%s/instmsg/aliases/%s"
                             "</d:href>"
                             "<r:description>"
                             "%s"
                             "</r:description>"
                             "</r:contact>"
                             "</r:notification-from>"
                             "<r:notification-to>"
                             "<r:contact>"
                             "<d:href>"
                             "http://%s/instmsg/aliases/%s"
                             "</d:href>"
                             "<r:description>"
                             "%s"
                             "</r:description>"
                             "</r:contact>"
                             "</r:notification-to>"
                             "%s"
                             "</r:%s>"
                             "</r:notification>",
                             action,
                             session->host,
                             session->user,
                             gaim_connection_get_display_name(hc->gc),
                             rvpb->host,
                             rvpb->user,
                             buddy->server_alias,
                             message,
                             action);

  hc->header = g_strdup_printf(
                               "%s %s HTTP/1.1\r\n"
                               "Host: %s\r\n"
                               "RVP-Notifications-Version: 0.2\r\n"
                               "RVP-Hop-Count: %d\r\n"
                               "%s"
                               "RVP-From-Principal: %s\r\n"
                               "Content-Type: text/xml\r\n"
                               "Content-length: %d\r\n",
			       hc->method,
			       hc->uri,
			       hc->host,
                               rvphopcount,
                               rvpacktype,
			       urlprincipal,
                               strlen(hc->body));
  hc->analyze = rvp_notify_cb;

  g_free(urlprincipal);

  if (gaim_proxy_connect(session->account, hc->host, RVP_PORT,
                         rvp_http_send, hc) != 0)
    {
      g_snprintf(buf, sizeof buf, _("buddy %s on host %s."),
                 gaim_buddy_get_name(buddy), hc->host);
      gaim_notify_error(gc, NULL, _("Unable To Connect"), buf);
    }

  return;
}

gchar *
generate_sessionid(const gchar *login)
{
  GaimCipher *cipher = NULL;
  GaimCipherContext *ctx = NULL;

  gchar *out = g_new0(gchar, 128);
  unsigned int output[4];
  time_t tim = time(NULL);
  GRand *grand = g_rand_new_with_seed(tim);
  int rnd = g_rand_int(grand);
  g_rand_free(grand);

  g_return_val_if_fail(login != NULL, NULL);
  g_return_val_if_fail(out != NULL, NULL);

  gaim_debug(GAIM_DEBUG_INFO, "rvp", "generate_sessionid %s\n",
             login);

  cipher = gaim_ciphers_find_cipher("md5");
  ctx = gaim_cipher_context_new(cipher, NULL);

  gaim_cipher_context_append(ctx, (const guchar *)login, strlen(login));
  gaim_cipher_context_append(ctx, (const guchar *)&tim, sizeof(tim));
  gaim_cipher_context_append(ctx, (const guchar *)&rnd, sizeof(rnd));
  gaim_cipher_context_digest(ctx, sizeof(output), (guchar *)output, NULL);

  gaim_cipher_context_destroy(ctx);
  sprintf(out, "{%08X-%04X-%04X-%04X-%04X%08X}", output[0], output[1]>>16, output[1]&0xFFFF, output[2]>>16, output[2]&0xFFFF, output[3]);

  return out;
}


void
rvp_invites_notify(RvpSession *session, GaimBuddy *buddy)
{
  RvpBuddy *rvpb = NULL;

  g_return_if_fail(session != NULL);
  g_return_if_fail(buddy != NULL);

  rvpb = buddy->proto_data;
  g_return_if_fail(rvpb != NULL);

  gaim_debug(GAIM_DEBUG_INFO, "rvp", "rvp_invites_notify %s\n",
             gaim_buddy_get_name(buddy));

  if (rvpb->sessionid == NULL)
    rvpb->sessionid = generate_sessionid(rvpb->user);

  gchar *msg = g_strdup_printf(
                               "<r:msgbody>"
                               "<r:mime-data>"
                               "<![CDATA[MIME-Version: 1.0\r\n"
                               "Content-Type: text/x-msmsgscontrol\r\n"
                               "Session-Id: %s\r\n"
                               "\r\n"
                               "\r\n"
                               "]]>"
                               "</r:mime-data>"
                               "</r:msgbody>",
                               rvpb->sessionid);

  rvp_notify(session, buddy, msg, 1, "RVP-Ack-Type: DeepOr\r\n");
  g_free(msg);

  return;
}


void
rvp_typing_notify(RvpSession *session, GaimBuddy *buddy)
{
  RvpBuddy *rvpb = NULL;

  g_return_if_fail(session != NULL);
  g_return_if_fail(buddy != NULL);

  rvpb = buddy->proto_data;
  g_return_if_fail(rvpb != NULL);

  gaim_debug(GAIM_DEBUG_INFO, "rvp", "rvp_typing_notify %s\n",
             gaim_buddy_get_name(buddy));

  if (rvpb->sessionid == NULL)
    {
      rvpb->sessionid = generate_sessionid(rvpb->user);
      rvp_invites_notify(session, buddy);
    }

  gchar *msg = g_strdup_printf(
                               "<r:msgbody>"
                               "<r:mime-data>"
                               "<![CDATA[MIME-Version: 1.0\r\n"
                               "Content-Type: text/x-msmsgscontrol\r\n"
                               "TypingUser: %s@%s\r\n"
                               "Session-Id: %s\r\n"
                               "\r\n"
                               "\r\n"
                               "]]>"
                               "</r:mime-data>"
                               "</r:msgbody>",
                               session->user,
                               session->host,
                               rvpb->sessionid);

  rvp_notify(session, buddy, msg, 1, "RVP-Ack-Type: DeepOr\r\n");
  g_free(msg);

  return;
}


void
rvp_send_notify(RvpSession *session, GaimBuddy *buddy, const gchar *message)
{
  RvpBuddy *rvpb = NULL;

  g_return_if_fail(session != NULL);
  g_return_if_fail(buddy != NULL);

  rvpb = buddy->proto_data;
  g_return_if_fail(rvpb != NULL);

  gaim_debug(GAIM_DEBUG_INFO, "rvp", "rvp_send_notify %s\n",
             gaim_buddy_get_name(buddy));

  if (rvpb->sessionid == NULL)
    {
      rvpb->sessionid = generate_sessionid(rvpb->user);
      rvp_invites_notify(session, buddy);
    }

  gchar *msg = g_strdup_printf(
                               "<r:msgbody>"
                               "<r:mime-data>"
                               "<![CDATA[MIME-Version: 1.0\r\n"
                               "Content-Type: text/plain; charset=UTF-8\r\n"
                               "X-MMS-IM-Format: FN=Tahoma; EF=; CO=0; CS=ee; PF=22\r\n"
                               "Session-Id: %s\r\n"
                               "\r\n"
                               "%s]]>"
                               "</r:mime-data>"
                               "</r:msgbody>",
                               rvpb->sessionid,
                               gaim_utf8_salvage(message));

  rvp_notify(session, buddy, msg, 1, "RVP-Ack-Type: DeepOr\r\n");
  g_free(msg);

  return;
}


void
rvp_send_request_file(RvpSession *session, GaimBuddy *buddy, GaimXfer *xfer)
{
  RvpBuddy *rvpb = NULL;
  RvpFile *rvpfile = NULL;
  GRand *grand = NULL;

  g_return_if_fail(session != NULL);
  g_return_if_fail(buddy != NULL);

  rvpb = buddy->proto_data;
  g_return_if_fail(rvpb != NULL);

  rvpfile = xfer->data;
  g_return_if_fail(rvpfile != NULL);

  gaim_debug(GAIM_DEBUG_INFO, "rvp", "rvp_send_request_file %s\n",
             gaim_buddy_get_name(buddy));

  if (rvpb->sessionid == NULL)
    {
      rvpb->sessionid = generate_sessionid(rvpb->user);
      rvp_invites_notify(session, buddy);
    }

  grand = g_rand_new_with_seed(time(NULL));
  rvpfile->cookie = g_rand_int_range(grand, 100000, 99999999);
  g_rand_free(grand);

  gchar *msg = g_strdup_printf(
                               "<r:msgbody>"
                               "<r:mime-data>"
                               "<![CDATA[MIME-Version: 1.0\r\n"
                               "Content-Type: text/x-msmsgsinvite; charset=UTF-8\r\n"
                               "Session-Id: %s\r\n"
                               "\r\n"
                               "Application-Name: File Transfer\r\n"
                               "Application-GUID: " RVP_GUID "\r\n"
                               "Invitation-Command: INVITE\r\n"
                               "Invitation-Cookie: %d\r\n"
                               "Application-File: %s\r\n"
                               "Application-FileSize: %d\r\n"
                               "\r\n"
                               "]]>"
                               "</r:mime-data>"
                               "</r:msgbody>",
                               rvpb->sessionid,
                               rvpfile->cookie,
                               gaim_xfer_get_filename(xfer),
                               gaim_xfer_get_size(xfer));

  rvp_notify(session, buddy, msg, 1, "RVP-Ack-Type: DeepOr\r\n");
  g_free(msg);

  return;
}


void
rvp_send_accept_file(RvpSession *session, GaimBuddy *buddy, GaimXfer *xfer)
{
  RvpBuddy *rvpb = NULL;
  RvpFile *rvpfile = NULL;

  g_return_if_fail(session != NULL);
  g_return_if_fail(buddy != NULL);

  rvpb = buddy->proto_data;
  g_return_if_fail(rvpb != NULL);

  rvpfile = xfer->data;
  g_return_if_fail(rvpfile != NULL);

  gaim_debug(GAIM_DEBUG_INFO, "rvp", "rvp_send_accept_file %s\n",
             gaim_buddy_get_name(buddy));

  if (rvpb->sessionid == NULL)
    {
      rvpb->sessionid = generate_sessionid(rvpb->user);
      rvp_invites_notify(session, buddy);
    }

  gchar *msg = g_strdup_printf(
                               "<r:msgbody>"
                               "<r:mime-data>"
                               "<![CDATA[MIME-Version: 1.0\r\n"
                               "Content-Type: text/x-msmsgsinvite; charset=UTF-8\r\n"
                               "Session-Id: %s\r\n"
                               "\r\n"
                               "Invitation-Command: ACCEPT\r\n"
                               "Invitation-Cookie: %d\r\n"
                               "Launch-Application: FALSE\r\n"
                               "Request-Data: IP-Address:\r\n"
                               "\r\n"
                               "]]>"
                               "</r:mime-data>"
                               "</r:msgbody>",
                               rvpb->sessionid,
                               rvpfile->cookie);

  rvp_notify(session, buddy, msg, 1, "RVP-Ack-Type: DeepOr\r\n");
  g_free(msg);

  return;
}

void
rvp_send_accept_info_file(RvpSession *session, GaimBuddy *buddy, GaimXfer *xfer)
{
  RvpBuddy *rvpb = NULL;
  RvpFile *rvpfile = NULL;
  GRand *grand = NULL;

  g_return_if_fail(session != NULL);
  g_return_if_fail(buddy != NULL);

  rvpb = buddy->proto_data;
  g_return_if_fail(rvpb != NULL);

  rvpfile = xfer->data;
  g_return_if_fail(rvpfile != NULL);

  gaim_debug(GAIM_DEBUG_INFO, "rvp", "rvp_send_accept_info_file %s\n",
             gaim_buddy_get_name(buddy));

  if (rvpb->sessionid == NULL)
    {
      rvpb->sessionid = generate_sessionid(rvpb->user);
      rvp_invites_notify(session, buddy);
    }

  grand = g_rand_new_with_seed(time(NULL));
  rvpfile->authcookie = g_rand_int_range(grand, 100000, 99999999);
  g_rand_free(grand);

  gchar *msg = g_strdup_printf(
                               "<r:msgbody>"
                               "<r:mime-data>"
                               "<![CDATA[MIME-Version: 1.0\r\n"
                               "Content-Type: text/x-msmsgsinvite; charset=UTF-8\r\n"
                               "Session-Id: %s\r\n"
                               "\r\n"
                               "Invitation-Command: ACCEPT\r\n"
                               "Invitation-Cookie: %d\r\n"
                               "IP-Address: %s\r\n"
                               "Port: %d\r\n"
                               "PortX: %d\r\n"
                               "AuthCookie: %d\r\n"
                               "Launch-Application: FALSE\r\n"
                               "Request-Data: IP-Address:\r\n"
                               "\r\n"
                               "]]>"
                               "</r:mime-data>"
                               "</r:msgbody>",
                               rvpb->sessionid,
                               rvpfile->cookie,
                               session->ip,
                               RVP_PORT1,
                               RVP_PORT2,
                               rvpfile->authcookie);

  rvp_notify(session, buddy, msg, 1, "RVP-Ack-Type: DeepOr\r\n");
  g_free(msg);

  return;
}


void
rvp_send_cancel_timeout_file(RvpSession *session, GaimBuddy *buddy, GaimXfer *xfer)
{
  RvpBuddy *rvpb = NULL;
  RvpFile *rvpfile = NULL;

  g_return_if_fail(session != NULL);
  g_return_if_fail(buddy != NULL);

  rvpb = buddy->proto_data;
  g_return_if_fail(rvpb != NULL);

  rvpfile = xfer->data;
  g_return_if_fail(rvpfile != NULL);

  gaim_debug(GAIM_DEBUG_INFO, "rvp", "rvp_send_cancel_timeout_file %s\n",
             gaim_buddy_get_name(buddy));

  if (rvpb->sessionid == NULL)
    {
      rvpb->sessionid = generate_sessionid(rvpb->user);
      rvp_invites_notify(session, buddy);
    }

  gchar *msg = g_strdup_printf(
                               "<r:msgbody>"
                               "<r:mime-data>"
                               "<![CDATA[MIME-Version: 1.0\r\n"
                               "Content-Type: text/x-msmsgsinvite; charset=UTF-8\r\n"
                               "Session-Id: %s\r\n"
                               "\r\n"
                               "Invitation-Command: CANCEL\r\n"
                               "Invitation-Cookie: %d\r\n"
                               "Cancel-Code: FTTIMEOUT\r\n"
                               "\r\n"
                               "]]>"
                               "</r:mime-data>"
                               "</r:msgbody>",
                               rvpb->sessionid,
                               rvpfile->cookie);

  rvp_notify(session, buddy, msg, 1, "RVP-Ack-Type: DeepOr\r\n");
  g_free(msg);

  return;
}
