/**
 * @file property.c Property functions and definitions
 *
 * gaim
 *
 * Gaim is the legal property of its developers, whose names are too numerous
 * to list here.  Please refer to the COPYRIGHT file distributed with this
 * source distribution.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include "internal.h"

#include "account.h"
#include "debug.h"
#include "xmlnode.h"

#include "session.h"
#include "httpconn.h"
#include "property.h"

void rvp_property_cb(RvpHttpconn *hc, const gchar *receive)
{
  GaimConnection *gc = NULL;

  gchar *tmp = NULL;

  g_return_if_fail(hc != NULL);
  g_return_if_fail(receive != NULL);

  gc = hc->gc;
  g_return_if_fail(gc != NULL);

  if (hc->inpa > 0)
    gaim_input_remove(hc->inpa);

  if (hc->fd > 0)
    close(hc->fd);

  if (strncmp(receive, "HTTP/1.1 207 Multi-Status\r\n", 27) == 0)
    {
      xmlnode *node = NULL;
      xmlnode *nodestatus = NULL;

      tmp = strstr(receive, "<?xml version=\"1.0\"?>");
      g_return_if_fail(tmp != NULL);

      node = xmlnode_from_str(tmp, strlen(tmp));
      g_return_if_fail(node != NULL);

      nodestatus = xmlnode_get_child(node,
				     "d:response/d:propstat/d:status");
      g_return_if_fail(nodestatus != NULL);

      if (strncmp(xmlnode_get_data(nodestatus),
		  "HTTP/1.1 200 Successful", 23) == 0)
        {
          gaim_debug(GAIM_DEBUG_INFO, "rvp", "rvp_property_cb success\n");

          nodestatus = xmlnode_get_child(node,
                                         "d:response/d:propstat/d:prop");
          g_return_if_fail(nodestatus != NULL);

          nodestatus = xmlnode_get_child(nodestatus, (gchar *)hc->option);
          g_return_if_fail(nodestatus != NULL);

          if (strcmp("d:displayname", (gchar *)hc->option) == 0)
            gaim_connection_set_display_name(gc, xmlnode_get_data(nodestatus));
        }
      else
        {
          gaim_debug(GAIM_DEBUG_ERROR, "rvp", "rvp_property_cb failed\n");
        }
      xmlnode_free(node);
    }

  return;
}

void
rvp_get_property(RvpSession *session, const gchar *login, const gchar *host,
                 char *property)
{
  GaimConnection *gc = NULL;

  RvpHttpconn *hc = g_new0(RvpHttpconn, 1);
  gchar buf[80];

  g_return_if_fail(session != NULL);
  g_return_if_fail(session->connected);
  g_return_if_fail(login != NULL);
  g_return_if_fail(host != NULL);
  g_return_if_fail(property != NULL);

  g_return_if_fail(hc != NULL);

  gc = gaim_account_get_connection(session->account);
  g_return_if_fail(gc != NULL);

  gaim_debug(GAIM_DEBUG_INFO, "rvp", "rvp_get_property\n");

  hc->gc = gc;
  hc->connection_step = 1;

  if (gc->inpa > 0)
    gaim_input_remove(gc->inpa);

  hc->method = "PROPFIND";
  hc->uri = g_strdup_printf(
			    "/instmsg/aliases/%s",
			    login);
  hc->callback = FALSE;
  hc->host = g_strdup(host);
  hc->option = (gpointer *)property;
  hc->body = g_strdup_printf(
                             "<?xml version=\"1.0\"?>\n"
                             "<d:propfind xmlns:d='DAV:' "
                             "xmlns:r='http://schemas.microsoft.com/rvp/'>"
                             "<d:prop>"
                             "<%s/>"
                             "</d:prop>"
                             "</d:propfind>",
                             property);

  hc->header = g_strdup_printf(
                               "%s %s HTTP/1.1\r\n"
                               "Depth: 0\r\n"
                               "RVP-Notifications-Version: 0.2\r\n"
                               "Host: %s\r\n"
                               "Content-Length: %d\r\n"
                               "Content-Type: text/xml\r\n"
                               "RVP-From-Principal: http://%s/instmsg/aliases/%s\r\n",
			       hc->method,
			       hc->uri,
                               hc->host,
                               strlen(hc->body),
                               session->host,
                               session->user);
  hc->analyze = rvp_property_cb;

  if (gaim_proxy_connect(session->account, hc->host, RVP_PORT,
                         rvp_http_send, hc) != 0)
    {
      g_snprintf(buf, sizeof buf, _("Connect to %s failed"), hc->host);
      gaim_connection_error(gc, buf);
    }

  return;
}

void
rvp_get_buddy_property(RvpSession *session, const GaimBuddy *buddy,
                       char *property)
{
  RvpBuddy *rvpb = NULL;

  g_return_if_fail(session != NULL);
  g_return_if_fail(buddy != NULL);
  g_return_if_fail(property != NULL);

  rvpb = buddy->proto_data;
  g_return_if_fail(rvpb != NULL);

  gaim_debug(GAIM_DEBUG_INFO, "rvp", "rvp_get_buddy_property %s\n",
             gaim_buddy_get_name(buddy));
  rvp_get_property(session, rvpb->user, rvpb->host, property);

  return;
}
