/**
 * @file rvp.c
 *
 * gaim - Rvp Protocol Plugin
 *
 * Copyright (C) 2005 Fabien Carrion <fabien.carrion@gmail.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include "internal.h"

#include "account.h"
#include "accountopt.h"
#include "blist.h"
#include "privacy.h"
#include "debug.h"
#include "network.h"
#include "notify.h"
#include "prpl.h"
#include "util.h"
#include "status.h"
#include "version.h"
#include "ft.h"

#include "session.h"
#include "state.h"
#include "property.h"
#include "subscribe.h"
#include "acl.h"
#include "msg.h"
#include "file.h"

#define RVP_HOST "localhost"

/**************************************************************************
 * Protocol Plugin ops
 **************************************************************************/

static const gchar *
rvp_list_icon(GaimAccount *a, GaimBuddy *b)
{
  return "msn";
}

static void
rvp_list_emblems(GaimBuddy *buddy, const gchar **se, const gchar **sw,
		 const gchar **nw, const gchar **ne)
{
  const gchar *emblems[4] = { NULL, NULL, NULL, NULL };
  GaimPresence *presence = NULL;
  int i = 0;

  g_return_if_fail(buddy != NULL);

  gaim_debug(GAIM_DEBUG_INFO, "rvp","rvp_list_emblems %s\n",
             gaim_buddy_get_name(buddy));

  presence = gaim_buddy_get_presence(buddy);

  if (!gaim_presence_is_online(presence))
    emblems[i++] = "offline";
  else if (gaim_presence_is_status_active(presence, "busy"))
    emblems[i++] = "occupied";
  else if (gaim_presence_is_status_active(presence, "brb") ||
	   gaim_presence_is_status_active(presence, "phone") ||
	   gaim_presence_is_status_active(presence, "lunch"))
    emblems[i++] = "extendedaway";
  else if (!gaim_presence_is_available(presence))
    emblems[i++] = "away";

  *se = emblems[0];
  *sw = emblems[1];
  *nw = emblems[2];
  *ne = emblems[3];

  return;
}

static gchar *
rvp_status_text(GaimBuddy *buddy)
{
  GaimPresence *presence = NULL;
  GaimStatus *status = NULL;

  g_return_val_if_fail(buddy != NULL, NULL);

  gaim_debug(GAIM_DEBUG_INFO, "rvp","rvp_status_text %s\n",
             gaim_buddy_get_name(buddy));

  presence = gaim_buddy_get_presence(buddy);
  status = gaim_presence_get_active_status(presence);

  if (!gaim_presence_is_available(presence) && !gaim_presence_is_idle(presence))
    return g_strdup(gaim_status_get_name(status));

  return NULL;
}


static gchar *
rvp_tooltip_text(GaimBuddy *buddy)
{
  GaimPresence *presence = NULL;
  GaimStatus *status = NULL;
  GString *s = NULL;

  g_return_val_if_fail(buddy != NULL, NULL);

  gaim_debug(GAIM_DEBUG_INFO, "rvp","rvp_tooltip_text %s\n",
             gaim_buddy_get_name(buddy));

  presence = gaim_buddy_get_presence(buddy);
  status = gaim_presence_get_active_status(presence);
  s = g_string_new("");

  if (gaim_presence_is_online(presence))
    {
      g_string_append_printf(s, _("\n<b>%s:</b> %s"), _("Status"),
			     gaim_presence_is_idle(presence) ?
			     _("Idle") : gaim_status_get_name(status));
    }

  return g_string_free(s, FALSE);
}


static GList *
rvp_status_types(GaimAccount *account)
{
  GaimStatusType *status = NULL;
  GList *types = NULL;

  g_return_val_if_fail(account != NULL, NULL);

  gaim_debug(GAIM_DEBUG_INFO, "rvp","rvp_status_types\n");

  status = gaim_status_type_new_full(GAIM_STATUS_OFFLINE,
                                     "offline", _("Offline"), FALSE, FALSE, FALSE);
  types = g_list_append(types, status);

  status = gaim_status_type_new_full(GAIM_STATUS_AVAILABLE,
                                     "available", _("Available"), FALSE, TRUE, FALSE);
  types = g_list_append(types, status);

  status = gaim_status_type_new_full(GAIM_STATUS_AVAILABLE,
                                     "idle", _("Idle"), FALSE, TRUE, FALSE);
  types = g_list_append(types, status);

  status = gaim_status_type_new_full(GAIM_STATUS_AWAY,
                                     "away", _("Away"), FALSE, TRUE, FALSE);
  types = g_list_append(types, status);

  status = gaim_status_type_new_full(GAIM_STATUS_AWAY,
                                     "brb", _("Be Right Back"), FALSE, TRUE, FALSE);
  types = g_list_append(types, status);

  status = gaim_status_type_new_full(GAIM_STATUS_UNAVAILABLE,
                                     "busy", _("Busy"), FALSE, TRUE, FALSE);
  types = g_list_append(types, status);

  status = gaim_status_type_new_full(GAIM_STATUS_UNAVAILABLE,
                                     "phone", _("On The Phone"), FALSE, TRUE, FALSE);
  types = g_list_append(types, status);

  status = gaim_status_type_new_full(GAIM_STATUS_AWAY,
                                     "lunch", _("Out To Lunch"), FALSE, TRUE, FALSE);
  types = g_list_append(types, status);

  status = gaim_status_type_new_full(GAIM_STATUS_HIDDEN,
                                     "hidden", _("Hidden"), FALSE, TRUE, FALSE);
  types = g_list_append(types, status);

  return types;
}

static void
rvp_login(GaimAccount *account, GaimStatus *status)
{
  GaimConnection *gc = NULL;
  RvpSession *session = NULL;
  const gchar *user = NULL;
  gchar *domain = NULL;

  g_return_if_fail(account != NULL);
  g_return_if_fail(status != NULL);

  gc = gaim_account_get_connection(account);
  g_return_if_fail(gc != NULL);

  user = gaim_account_get_username(account);
  g_return_if_fail(user != NULL);

  gaim_debug(GAIM_DEBUG_INFO, "rvp","rvp_login\n");

  session = rvp_session_new(account);
  g_return_if_fail(session != NULL);

  gc->proto_data = session;

  domain = strchr(user, '/');
  g_return_if_fail(domain != NULL);

  session->host     = g_strdup(gaim_account_get_string(account, "server", RVP_HOST));

  session->user     = g_strndup(user, domain - user);
  session->password = g_strdup(gaim_account_get_password(gc->account));

  session->domain   = g_strdup(domain + 1);

  rvp_session_login(session, gc);

  return;
}

static void
rvp_close(GaimConnection *gc)
{
  RvpSession *session = NULL;

  GaimBlistNode *gnode = NULL;
  GaimBlistNode *cnode = NULL;
  GaimBlistNode *bnode = NULL;

  g_return_if_fail(gc != NULL);

  session = gc->proto_data;
  g_return_if_fail(session != NULL);

  gaim_debug(GAIM_DEBUG_INFO, "rvp","rvp_close\n");

  for(gnode = gaim_get_blist()->root; gnode; gnode = gnode->next)
    {
      if(!GAIM_BLIST_NODE_IS_GROUP(gnode))
        continue;
      for(cnode = gnode->child; cnode; cnode = cnode->next)
        {
          if(!GAIM_BLIST_NODE_IS_CONTACT(cnode))
            continue;
          for(bnode = cnode->child; bnode; bnode = bnode->next)
            {
              if(!GAIM_BLIST_NODE_IS_BUDDY(bnode))
                continue;
              if(((GaimBuddy*)bnode)->account == gc->account)
                rvp_unsubscribe(session, (GaimBuddy*)bnode);
            }
        }
    }

  rvp_change_status(session, RVP_OFFLINE);
  rvp_session_logout(session);
  gc->proto_data = NULL;

  return;
}

static int
rvp_send_im(GaimConnection *gc, const char *name, const char *message, GaimConvImFlags flags)
{
  RvpSession *session = NULL;
  GaimBuddy *buddy = NULL;

  g_return_val_if_fail(gc != NULL, -1);
  g_return_val_if_fail(name != NULL, -1);
  g_return_val_if_fail(message != NULL, -1);

  session = gc->proto_data;
  g_return_val_if_fail(session != NULL, -1);

  buddy = gaim_find_buddy(session->account, name);
  g_return_val_if_fail(buddy != NULL, -1);

  gaim_debug(GAIM_DEBUG_INFO, "rvp","rvp_send_im to %s\n",
             name);

  rvp_send_notify(session, buddy, message);

  return 1;
}

static int
rvp_typing(GaimConnection *gc, const char *name, int typing)
{
  RvpSession *session = NULL;
  GaimBuddy *buddy = NULL;

  g_return_val_if_fail(gc != NULL, -1);
  g_return_val_if_fail(name != NULL, -1);

  session = gc->proto_data;
  g_return_val_if_fail(session != NULL, -1);

  buddy = gaim_find_buddy(session->account, name);
  g_return_val_if_fail(buddy != NULL, -1);

  gaim_debug(GAIM_DEBUG_INFO, "rvp","rvp_typing to %s\n",
             name);

  if (typing == GAIM_TYPING)
    rvp_typing_notify(session, buddy);

  return 1;
}


static void
rvp_set_status(GaimAccount *account, GaimStatus *status)
{
  GaimConnection *gc = NULL;
  RvpSession *session = NULL;
  const char *state = NULL;
  int rvpstatus = -1;

  g_return_if_fail(account != NULL);
  g_return_if_fail(status != NULL);

  gc = gaim_account_get_connection(account);
  if (gc != NULL)
    session = gc->proto_data;

  state = gaim_status_get_id(status);

  gaim_debug(GAIM_DEBUG_INFO, "rvp", "rvp_set_status %s\n",
             gaim_status_get_name(status));

  if (!strcmp(state, "offline") && (gc != NULL))
    {
      gaim_account_disconnect(account);
      return;
    }
  else if (strcmp(state, "offline") && (gc == NULL))
    {
      gaim_account_connect(account);
      return;
    }
  else if (!strcmp(state, "away"))
    rvpstatus = RVP_AWAY;
  else if (!strcmp(state, "brb"))
    rvpstatus = RVP_BRB;
  else if (!strcmp(state, "busy"))
    rvpstatus = RVP_BUSY;
  else if (!strcmp(state, "phone"))
    rvpstatus = RVP_PHONE;
  else if (!strcmp(state, "lunch"))
    rvpstatus = RVP_LUNCH;
  else if (!strcmp(state, "hidden"))
    rvpstatus = RVP_HIDDEN;
  else if (0) /* how do we detect idle with new status? */
    rvpstatus = RVP_IDLE;
  else
    rvpstatus = RVP_ONLINE;
  if (gc)
    rvp_change_status(session, rvpstatus);

  return;
}

static void
rvp_set_idle(GaimConnection *gc, int time)
{
  g_return_if_fail(gc != NULL);

  gaim_debug(GAIM_DEBUG_INFO, "rvp","rvp_set_idle %d.\n",
             time);

  rvp_change_status(gc->proto_data, time ? RVP_IDLE : RVP_ONLINE);

  return;
}

static void
rvp_add_buddy(GaimConnection *gc, GaimBuddy *buddy, GaimGroup *group)
{
  RvpSession *session = NULL;
  const gchar *name = NULL;
  GaimBlistNode *gnode = NULL;
  GaimBlistNode *cnode = NULL;
  GaimBlistNode *bnode = NULL;
  int list = 0;

  g_return_if_fail(gc != NULL);
  g_return_if_fail(buddy != NULL);
  g_return_if_fail(group != NULL);

  session = gc->proto_data;
  g_return_if_fail(session != NULL);

  name = gaim_buddy_get_name(buddy);
  g_return_if_fail(name != NULL);

  if (!gaim_email_is_valid(name))
    {
      gchar *buf = g_strdup_printf(_("buddy %s is not an email."),
                                   name);
      gaim_notify_error(gc, NULL, _("Unable To Add"), buf);
      g_free(buf);
      gaim_blist_remove_buddy(buddy);
      return;
    }

  for(gnode = gaim_get_blist()->root; gnode; gnode = gnode->next)
    {
      if(!GAIM_BLIST_NODE_IS_GROUP(gnode))
        continue;
      for(cnode = gnode->child; cnode; cnode = cnode->next)
        {
          if(!GAIM_BLIST_NODE_IS_CONTACT(cnode))
            continue;
          for(bnode = cnode->child; bnode; bnode = bnode->next)
            {
              if(!GAIM_BLIST_NODE_IS_BUDDY(bnode))
                continue;
              if(((GaimBuddy*)bnode)->account == gc->account)
                {
                  if (strcmp(gaim_buddy_get_name((GaimBuddy*)bnode),
                             name) == 0)
                    list++;
                  if (list > 1)
                    {
                      gchar *buf = g_strdup_printf(_("buddy %s already in internal list."),
                                                   name);
                      gaim_notify_error(gc, NULL, _("Unable To Add"), buf);
                      g_free(buf);
                      gaim_blist_remove_buddy(buddy);
                      return;
                    }
                }
            }
        }
    }

  gaim_debug(GAIM_DEBUG_INFO, "rvp","rvp_add_buddy %s\n",
             name);

  rvp_subscribe(session, buddy);
  rvp_set_acl_permit(session, buddy);

  return;
}

static void
rvp_remove_buddy(GaimConnection *gc, GaimBuddy *buddy, GaimGroup *group)
{
  RvpSession *session = NULL;

  g_return_if_fail(gc != NULL);
  g_return_if_fail(buddy != NULL);
  g_return_if_fail(group != NULL);

  session = gc->proto_data;
  g_return_if_fail(session != NULL);

  gaim_debug(GAIM_DEBUG_INFO, "rvp","rvp_remove_buddy %s\n",
             gaim_buddy_get_name(buddy));

  rvp_unsubscribe(session, buddy);

  return;
}

static void
rvp_add_permit(GaimConnection *gc, const char *who)
{
  RvpSession *session = NULL;

  g_return_if_fail(gc != NULL);
  g_return_if_fail(who != NULL);

  session = gc->proto_data;
  g_return_if_fail(session != NULL);

  gaim_debug(GAIM_DEBUG_INFO, "rvp","rvp_add_permit %s\n",
             who);

  rvp_set_acl_permit(session, gaim_find_buddy(session->account, who));

  return;
}


static void
rvp_add_deny(GaimConnection *gc, const char *who)
{
  RvpSession *session = NULL;

  g_return_if_fail(gc != NULL);
  g_return_if_fail(who != NULL);

  session = gc->proto_data;
  g_return_if_fail(session != NULL);

  gaim_debug(GAIM_DEBUG_INFO, "rvp","rvp_add_deny %s\n",
             who);

  rvp_set_acl_deny(session, gaim_find_buddy(session->account, who));

  return;
}

static void
rvp_set_permit_deny(GaimConnection *gc)
{
  GaimAccount *account = NULL;
  GSList *who = NULL;

  g_return_if_fail(gc != NULL);

  account = gc->account;
  g_return_if_fail(account != NULL);

  gaim_debug(GAIM_DEBUG_INFO, "rvp","rvp_set_permit_deny %d\n",
             account->perm_deny);

  for (who = account->deny;who;who = who->next)
    rvp_add_deny(gc, who->data);
  for (who = account->permit;who;who = who->next)
    rvp_add_permit(gc, who->data);

  return;
}


static void
rvp_keep_alive(GaimConnection *gc)
{
  RvpSession *session = NULL;

  g_return_if_fail(gc != NULL);

  session = gc->proto_data;
  g_return_if_fail(session != NULL);

  gaim_debug(GAIM_DEBUG_INFO, "rvp","rvp_keep_alive\n");
  rvp_session_login(session, gc);

  return;
}

static void
rvp_buddy_free(GaimBuddy *buddy)
{
  RvpBuddy *b = NULL;

  g_return_if_fail(buddy != NULL);

  b = buddy->proto_data;
  g_return_if_fail(b != NULL);

  gaim_debug(GAIM_DEBUG_INFO, "rvp","rvp_buddy_free %s\n",
             gaim_buddy_get_name(buddy));

  g_free(b->host);
  g_free(b->user);
  g_free(b->sessionid);
  g_free(b);

  buddy->proto_data = NULL;

  return;
}

static gboolean
rvp_can_receive_file(GaimConnection *gc, const gchar *who)
{
  GaimAccount *account = NULL;
  gchar *normal = NULL;
  gboolean ret = FALSE;

  g_return_val_if_fail(gc != NULL, FALSE);
  g_return_val_if_fail(who != NULL, FALSE);

  account = gaim_connection_get_account(gc);
  g_return_val_if_fail(account != NULL, FALSE);

  normal = g_strdup(gaim_normalize_nocase(account, gaim_account_get_username(account)));

  gaim_debug(GAIM_DEBUG_INFO, "rvp","rvp_can_receive_file %s\n",
             who);

  ret = strcmp(normal, gaim_normalize_nocase(account, who));
  g_free(normal);

  return ret;
}


static void
rvp_send_file(GaimConnection *gc, const char *who, const char *filename)
{
  RvpSession *session = NULL;
  GaimXfer *xfer = NULL;
  RvpFile *file = rvp_file_new();

  session = gc->proto_data;
  g_return_if_fail(session != NULL);

  gaim_debug(GAIM_DEBUG_INFO, "rvp", "rvp_send_file %s to %s\n",
             filename, who);

  /* Build the file transfer handle */
  xfer = gaim_xfer_new(gc->account, GAIM_XFER_SEND, who);
  g_return_if_fail(xfer != NULL);

  xfer->data = file;
  g_return_if_fail(file != NULL);

  /* Setup our I/O op functions */
  gaim_xfer_set_init_fnc(xfer, rvp_xfer_init);
  gaim_xfer_set_start_fnc(xfer, rvp_xfer_start_send);
  gaim_xfer_set_end_fnc(xfer, rvp_xfer_end);
  gaim_xfer_set_write_fnc(xfer, rvp_xfer_write);
  gaim_xfer_set_request_denied_fnc(xfer, rvp_xfer_cancel);
  gaim_xfer_set_cancel_send_fnc(xfer, rvp_xfer_cancel_notify);

  /* Now perform the request */
  if (filename)
    gaim_xfer_request_accepted(xfer, filename);
  else
    gaim_xfer_request(xfer);

  return;
}


static GaimPluginProtocolInfo prpl_info =
{
  0,
  NULL,					/* user_splits */
  NULL,					/* protocol_options */
  {"png", 0, 0, 96, 96, GAIM_ICON_SCALE_SEND},        /* icon_spec */
  rvp_list_icon,         /* list_icon */
  rvp_list_emblems,      /* list_emblems */
  rvp_status_text,       /* status_text */
  rvp_tooltip_text,      /* tooltip_text */
  rvp_status_types,      /* away_states */
  NULL,                  /* blist_node_menu */
  NULL,                  /* chat_info */
  NULL,                  /* chat_info_defaults */
  rvp_login,             /* login */
  rvp_close,             /* close */
  rvp_send_im,           /* send_im */
  NULL,                  /* set_info */
  rvp_typing,            /* send_typing */
  NULL,                  /* get_info */
  rvp_set_status,        /* set_status */
  rvp_set_idle,          /* set_idle */
  NULL,                  /* change_passwd */
  rvp_add_buddy,         /* add_buddy */
  NULL,                  /* add_buddies */
  rvp_remove_buddy,      /* remove_buddy */
  NULL,                  /* remove_buddies */
  rvp_add_permit,        /* add_permit */
  rvp_add_deny,          /* add_deny */
  rvp_add_deny,          /* rem_permit */
  rvp_add_permit,        /* rem_deny */
  rvp_set_permit_deny,   /* set_permit_deny */
  NULL,                  /* join_chat */
  NULL,                  /* reject chat invite */
  NULL,                  /* get_chat_name */
  NULL,                  /* chat_invite */
  NULL,                  /* chat_leave */
  NULL,                  /* chat_whisper */
  NULL,                  /* chat_send */
  rvp_keep_alive,        /* keepalive */
  NULL,                  /* register_user */
  NULL,                  /* get_cb_info */
  NULL,                  /* get_cb_away */
  NULL,                  /* alias_buddy */
  NULL,                  /* group_buddy */
  NULL,                  /* rename_group */
  rvp_buddy_free,        /* buddy_free */
  NULL,                  /* convo_closed */
  gaim_normalize_nocase, /* normalize */
  NULL,                  /* set_buddy_icon */
  NULL,                  /* remove_group */
  NULL,                  /* get_cb_real_name */
  NULL,                  /* set_chat_topic */
  NULL,                  /* find_blist_chat */
  NULL,                  /* roomlist_get_list */
  NULL,                  /* roomlist_cancel */
  NULL,                  /* roomlist_expand_category */
  rvp_can_receive_file,  /* can_receive_file */
  rvp_send_file          /* send_file */
};

static GaimPluginInfo info =
{
  GAIM_PLUGIN_MAGIC,
  GAIM_MAJOR_VERSION,
  GAIM_MINOR_VERSION,
  GAIM_PLUGIN_PROTOCOL,                             /**< type           */
  NULL,                                             /**< ui_requirement */
  0,                                                /**< flags          */
  NULL,                                             /**< dependencies   */
  GAIM_PRIORITY_DEFAULT,                            /**< priority       */

  "prpl-rvp",                                       /**< id             */
  "RVP",                                            /**< name           */
  VERSION,                                          /**< version        */
                                                    /**  summary        */
  N_("RVP Protocol Plugin"),
                                                    /**  description    */
  N_("RVP Protocol Plugin"),
  N_("Fabien Carrion <fabien.carrion@gmail.com>"),  /**< author         */
  GAIM_WEBSITE,                                     /**< homepage       */

  NULL,                                             /**< load           */
  NULL,                                             /**< unload         */
  NULL,                                             /**< destroy        */

  NULL,                                             /**< ui_info        */
  &prpl_info,                                       /**< extra_info     */
  NULL,                                             /**< prefs_info     */
  NULL
};

static void
init_plugin(GaimPlugin *plugin)
{
  GaimAccountUserSplit *split = NULL;
  GaimAccountOption *option = NULL;

  g_return_if_fail(plugin != NULL);

  split = gaim_account_user_split_new(_("Domain"), "", '/');
  prpl_info.user_splits = g_list_append(prpl_info.user_splits, split);

  option = gaim_account_option_string_new(_("Login server"), "server",
                                          RVP_HOST);
  prpl_info.protocol_options = g_list_append(prpl_info.protocol_options,
                                             option);
}

GAIM_INIT_PLUGIN(rvp, init_plugin, info);
