/**
 * @file session.c RVP session functions
 *
 * gaim
 *
 * Copyright (C) 2005 Fabien Carrion <fabien.carrion@gmail.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include "internal.h"

#include "account.h"
#include "network.h"
#include "proxy.h"
#include "debug.h"
#include "util.h"

#include "session.h"
#include "httpconn.h"
#include "state.h"
#include "property.h"
#include "acl.h"
#include "incomconn.h"

RvpSession *
rvp_session_new(GaimAccount *account)
{
  RvpSession *session = NULL;

  g_return_val_if_fail(account != NULL, NULL);

  session = g_new0(RvpSession, 1);

  session->account = account;

  session->host = NULL;
  session->domain = NULL;
  session->user = NULL;
  session->password = NULL;
  session->ip = NULL;

  session->connected = FALSE;

  session->state = RVP_OFFLINE;
  session->modstate = 0;
  session->state_lifetime = 0;
  session->fd = 0;
  session->listenfd = 0;
  session->subscriptionid = 0;
  session->subscriptionlifetime = 14400;

  return session;
}

void
rvp_login_cb(RvpHttpconn *hc, const gchar *receive)
{
  GaimConnection *gc = NULL;
  RvpSession *session = NULL;
  gchar buf[80];

  g_return_if_fail(hc != NULL);
  g_return_if_fail(receive != NULL);

  gc = hc->gc;
  g_return_if_fail(gc != NULL);

  session = gc->proto_data;
  g_return_if_fail(session != NULL);

  if (hc->inpa > 0)
    gaim_input_remove(hc->inpa);

  if (hc->fd > 0)
    close(hc->fd);

  if (strncmp(receive, "HTTP/1.1 400 Bad Request\r\n", 26) == 0)
    {
      gaim_connection_error(gc, _("Authentication failed"));
      session->connected = FALSE;
      return;
    }

  if (strncmp(receive, "HTTP/1.1 401 Access Denied\r\n", 28) == 0)
    {
      gaim_connection_error(gc, _("Authentication failed"));
      session->connected = FALSE;
      return;
    }

  if (strncmp(receive, "HTTP/1.1 200 Successful\r\n", 25) == 0)
    {
      gchar *tmp = NULL;

      if ((tmp = getHeader(receive, "Subscription-Id: ")) == NULL)
	{
	  gaim_connection_error(gc, _("Subscription Id unknown"));
	  return;
	}
      session->subscriptionid = atoi(tmp);
      g_free(tmp);

      if ((tmp = getHeader(receive, "Subscription-Lifetime: ")) == NULL)
	{
          session->subscriptionlifetime = 14400;
	}
      else
        {
          session->subscriptionlifetime = atoi(tmp);
          g_free(tmp);
        }

      gaim_debug(GAIM_DEBUG_INFO, "rvp", "rvp_login_cb success\n");

      if (session->connected == FALSE)
        {
          g_snprintf(buf, sizeof(buf), _("Signed: %s"), session->user);
          gaim_connection_update_progress(gc, buf, hc->connection_step, RVP_CONNECT_STEPS);

          gaim_connection_set_state(gc, GAIM_CONNECTED);

          session->connected = TRUE;
          rvp_change_status(session, RVP_ONLINE);
          rvp_get_property(session, session->user, session->host, "d:displayname");
          rvp_get_acl_buddies(gc);
        }
    }

  return;
}

void
rvp_session_login(RvpSession *session, GaimConnection *gc)
{
  RvpHttpconn *hc = g_new0(RvpHttpconn, 1);
  gchar buf[80];

  gchar *tmp = NULL;

  g_return_if_fail(session != NULL);
  g_return_if_fail(gc != NULL);

  g_return_if_fail(hc != NULL);

  gaim_debug(GAIM_DEBUG_INFO, "rvp", "rvp_session_login\n");

  hc->gc = gc;
  hc->connection_step = 1;

  if (gc->inpa > 0)
    gaim_input_remove(gc->inpa);

  if (session->connected == FALSE)
    {
      g_snprintf(buf, sizeof buf, _("Looking up %s"), session->host);
      gaim_connection_update_progress(gc, buf, hc->connection_step, RVP_CONNECT_STEPS);
      gaim_connection_set_state(gc, GAIM_CONNECTING);

      session->fd = gaim_network_listen_range(1035, 1135);
      if (session->fd == -1)
        {
          gaim_connection_error(gc, _("Could not create listen socket"));
          return;
        }
      session->listenfd = gaim_input_add(session->fd, GAIM_INPUT_READ, rvp_newconn_cb, gc);
    }

  switch (session->subscriptionid)
    {
    case -1:
      session->subscriptionid = 0;
      return;
    case 0:
      tmp = g_strdup("Notification-Type: pragma/notify");
      hc->callback = TRUE;
      break;
    default:
      tmp = g_strdup_printf(
                            "Subscription-Id: %d",
                            session->subscriptionid);
      hc->callback = FALSE;
      break;
    }

  hc->method = "SUBSCRIBE";
  hc->uri = g_strdup_printf(
			    "/instmsg/aliases/%s",
			    session->user);
  hc->host = g_strdup(session->host);
  hc->option = NULL;
  hc->body = g_strdup("");

  hc->header = g_strdup_printf(
			       "%s %s HTTP/1.1\r\n"
			       "Subscription-Lifetime: %d\r\n"
			       "%s\r\n"
			       "User-Agent: MSMSGS\r\n"
			       "Host: %s\r\n"
			       "RVP-Notifications-Version: 0.2\r\n"
			       "RVP-From-Principal: http://%s/instmsg/aliases/%s\r\n"
			       "Content-Length: 0\r\n",
			       hc->method,
			       hc->uri,
                               session->subscriptionlifetime,
                               tmp,
			       hc->host,
			       session->host,
			       session->user);
  hc->analyze = rvp_login_cb;

  g_free(tmp);
  if (gaim_proxy_connect(session->account, hc->host, RVP_PORT,
                         rvp_http_send, hc) != 0)
    {
      g_snprintf(buf, sizeof buf, _("Connect to %s failed"), hc->host);
      gaim_connection_error(gc, buf);
    }

  return;
}

void
rvp_logout_cb(RvpHttpconn *hc, const gchar *receive)
{
  GaimConnection *gc = NULL;
  RvpSession *session = NULL;

  g_return_if_fail(hc != NULL);
  g_return_if_fail(receive != NULL);

  gc = hc->gc;
  g_return_if_fail(gc != NULL);

  session = gc->proto_data;
  g_return_if_fail(session != NULL);

  if (hc->inpa > 0)
    gaim_input_remove(hc->inpa);

  if (hc->fd > 0)
    close(hc->fd);

  if (strncmp(receive, "HTTP/1.1 200 Successful\r\n", 25) == 0)
    gaim_debug(GAIM_DEBUG_INFO, "rvp", "rvp_logout_cb success\n");
  else
    gaim_debug(GAIM_DEBUG_INFO, "rvp", "rvp_logout_cb failed\n");

  gaim_connection_set_state(gc, GAIM_DISCONNECTED);

  return;
}

void
rvp_session_logout(RvpSession *session)
{
  GaimConnection *gc = NULL;

  RvpHttpconn *hc = g_new0(RvpHttpconn, 1);
  gchar buf[80];

  g_return_if_fail(session != NULL);
  g_return_if_fail(session->connected);

  gc = gaim_account_get_connection(session->account);
  g_return_if_fail(gc != NULL);

  gaim_debug(GAIM_DEBUG_INFO, "rvp", "rvp_session_logout\n");

  hc->gc = gc;
  hc->connection_step = 1;

  if (gc->inpa > 0)
    gaim_input_remove(gc->inpa);

  hc->method = "UNSUBSCRIBE";
  hc->uri = g_strdup_printf(
			    "/instmsg/aliases/%s",
			    session->user);
  hc->callback = FALSE;
  hc->host = g_strdup(session->host);
  hc->option = NULL;
  hc->body = g_strdup("");

  hc->header = g_strdup_printf(
                               "%s %s HTTP/1.1\r\n"
                               "Host: %s\r\n"
                               "RVP-Notifications-Version: 0.2\r\n"
                               "RVP-From-Principal: http://%s/instmsg/aliases/%s\r\n"
                               "Subscription-Id: %d\r\n"
                               "Content-Length: 0\r\n",
			       hc->method,
                               hc->uri,
                               hc->host,
                               session->host,
                               session->user,
                               session->subscriptionid);
  hc->analyze = rvp_logout_cb;

  session->connected = FALSE;
  session->subscriptionid = -1;

  if (session->modstate)
    gaim_timeout_remove(session->modstate);

  if (session->listenfd)
    gaim_input_remove(session->listenfd);

  if (session->fd > 0)
    {
      close(session->fd);
      session->fd = -1;
    }

  if (gaim_proxy_connect(session->account, hc->host, RVP_PORT,
                         rvp_http_send, hc) != 0)
    {
      g_snprintf(buf, sizeof buf, _("Connect to %s failed"), hc->host);
      gaim_connection_error(gc, buf);
    }

  return;
}
