/**
 * @file state.c State functions and definitions
 *
 * gaim
 *
 * Gaim is the legal property of its developers, whose names are too numerous
 * to list here.  Please refer to the COPYRIGHT file distributed with this
 * source distribution.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include "internal.h"

#include "account.h"
#include "debug.h"
#include "xmlnode.h"

#include "session.h"
#include "httpconn.h"
#include "state.h"

static const gchar *away_text[] =
{
	N_("available"),
	N_("busy"),
	N_("idle"),
	N_("brb"),
	N_("away"),
	N_("phone"),
	N_("lunch"),
	N_("offline"),
	N_("hidden")
};


gboolean
rvp_keep_status(RvpSession *session)
{
  g_return_val_if_fail(session != NULL, FALSE);
  g_return_val_if_fail(session->connected != FALSE, FALSE);

  gaim_debug(GAIM_DEBUG_INFO, "rvp", "rvp_keep_status\n");

  rvp_change_status(session, session->state);

  return TRUE;
}


void rvp_status_cb(RvpHttpconn *hc, const gchar *receive)
{
  GaimConnection *gc = NULL;
  RvpSession *session = NULL;
  xmlnode *node = NULL;
  xmlnode *nodestatus = NULL;
  gint state_lifetime = 1200;
  gchar *tmp = NULL;

  g_return_if_fail(hc != NULL);
  g_return_if_fail(receive != NULL);

  gc = hc->gc;
  g_return_if_fail(gc != NULL);

  session = gc->proto_data;
  g_return_if_fail(session != NULL);

  if (hc->inpa > 0)
    gaim_input_remove(hc->inpa);

  if (hc->fd > 0)
    close(hc->fd);

  if (strncmp(receive, "HTTP/1.1 207 Multi-Status\r\n", 27) == 0)
    {
      tmp = strstr(receive, "<?xml version=\"1.0\"?>");
      g_return_if_fail(tmp != NULL);

      node = xmlnode_from_str(tmp, strlen(tmp));
      g_return_if_fail(node != NULL);

      nodestatus = xmlnode_get_child(node,
				     "d:response/d:propstat/d:status");
      g_return_if_fail(nodestatus != NULL);

      if (strncmp(xmlnode_get_data(nodestatus),
		  "HTTP/1.1 200 Successful", 23) == 0)
	gaim_debug(GAIM_DEBUG_INFO, "rvp", "rvp_status_cb success\n");
      else
        gaim_debug(GAIM_DEBUG_ERROR, "rvp", "rvp_status_cb failed\n");

      nodestatus = xmlnode_get_child(node,
				     "d:response/d:propstat/d:prop/r:state/r:leased-value/d:timeout");
      g_return_if_fail(nodestatus != NULL);
      state_lifetime = atoi(xmlnode_get_data(nodestatus));

      xmlnode_free(node);

      if (session->state_lifetime != state_lifetime)
        {
          session->state_lifetime = state_lifetime;
          gaim_timeout_remove(session->modstate);
          session->modstate = gaim_timeout_add((session->state_lifetime - 2) * 10,
                                               (GSourceFunc) rvp_keep_status,
                                               session);
        }
    }

  return;
}

void
rvp_change_status(RvpSession *session, RvpAwayType state)
{
  GaimConnection *gc = NULL;

  RvpHttpconn *hc = g_new0(RvpHttpconn, 1);
  gchar buf[80];

  const gchar *state_text;

  g_return_if_fail(session != NULL);
  g_return_if_fail(session->connected);

  g_return_if_fail(hc != NULL);

  gc = gaim_account_get_connection(session->account);
  g_return_if_fail(gc != NULL);

  state_text = rvp_state_get_text(state);
  g_return_if_fail(state_text != NULL);

  session->state = state;

  gaim_debug(GAIM_DEBUG_INFO, "rvp", "rvp_change_status %s\n", state_text);

  hc->gc = gc;
  hc->connection_step = 1;

  if (gc->inpa > 0)
    gaim_input_remove(gc->inpa);

  hc->method = "PROPPATCH";
  hc->uri = g_strdup_printf(
			    "/instmsg/aliases/%s",
			    session->user);
  hc->callback = FALSE;
  hc->host = g_strdup(session->host);
  hc->option = NULL;
  hc->body = g_strdup_printf(
                             "<?xml version=\"1.0\"?>\n"
                             "<d:propertyupdate xmlns:d='DAV:' "
                             "xmlns:r='http://schemas.microsoft.com/rvp/' "
                             "xmlns:a='http://schemas.microsoft.com/rvp/acl/'>"
                             "<d:set>"
                             "<d:prop>"
                             "<r:state>"
                             "<r:leased-value>"
                             "<r:value>"
                             "%s"
                             "</r:value>"
                             "<r:default-value>"
                             "<r:offline/>"
                             "</r:default-value>"
                             "<d:timeout>1200</d:timeout>"
                             "</r:leased-value>"
                             "</r:state>"
                             "</d:prop>"
                             "</d:set>"
                             "</d:propertyupdate>",
                             state_text);

  hc->header = g_strdup_printf(
                               "%s %s HTTP/1.1\r\n"
                               "RVP-Notifications-Version: 0.2\r\n"
                               "Host: %s\r\n"
                               "Content-Length: %d\r\n"
                               "Content-Type: text/xml\r\n"
                               "RVP-From-Principal: http://%s/instmsg/aliases/%s\r\n",
			       hc->method,
                               hc->uri,
                               hc->host,
                               strlen(hc->body),
                               session->host,
                               session->user);
  hc->analyze = rvp_status_cb;

  if (gaim_proxy_connect(session->account, hc->host, RVP_PORT,
                         rvp_http_send, hc) != 0)
    {
      g_snprintf(buf, sizeof buf, _("Connect to %s failed"), hc->host);
      gaim_connection_error(gc, buf);
    }

  return;
}

const gchar *
rvp_away_get_text(RvpAwayType type)
{
  g_return_val_if_fail(type <= RVP_HIDDEN, NULL);

  return _(away_text[type]);
}

const gchar *
rvp_state_get_text(RvpAwayType state)
{
  static char *status_text[] =
    {
      "<r:online/>", "<r:busy/>", "<r:idle/>",
      "<r:back-soon/>", "<r:away/>", "<r:on-phone/>",
      "<r:at-lunch/>", "<r:offline/>", "<r:invisible/>"
    };

  return status_text[state];
}

int
rvp_state_get_number(const char *state)
{
  g_return_val_if_fail(state != NULL, RVP_OFFLINE);

  if (!g_ascii_strcasecmp(state, "<r:online/>"))
    return RVP_ONLINE;
  else if (!g_ascii_strcasecmp(state, "<r:busy/>"))
    return RVP_BUSY;
  else if (!g_ascii_strcasecmp(state, "<r:idle/>"))
    return RVP_IDLE;
  else if (!g_ascii_strcasecmp(state, "<r:back-soon/>"))
    return RVP_BRB;
  else if (!g_ascii_strcasecmp(state, "<r:away/>"))
    return RVP_AWAY;
  else if (!g_ascii_strcasecmp(state, "<r:on-phone/>"))
    return RVP_PHONE;
  else if (!g_ascii_strcasecmp(state, "<r:at-lunch/>"))
    return RVP_LUNCH;
  else if (!g_ascii_strcasecmp(state, "<r:offline/>"))
    return RVP_OFFLINE;
  else if (!g_ascii_strcasecmp(state, "<r:invisible/>"))
    return RVP_HIDDEN;

  return RVP_OFFLINE;
}
