/**
 * @file subscribe.c RVP subscribe buddy functions
 *
 * gaim
 *
 * Copyright (C) 2005 Fabien Carrion <fabien.carrion@gmail.com>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include "internal.h"

#include "account.h"
#include "debug.h"
#include "xmlnode.h"
#include "notify.h"

#include "session.h"
#include "httpconn.h"
#include "state.h"
#include "subscribe.h"

gboolean rvp_resubscribe(GaimBuddy *buddy);

void
rvp_subscribe_cb(RvpHttpconn *hc, const gchar *receive)
{
  GaimConnection *gc = NULL;
  RvpSession *session = NULL;
  GaimBuddy *b = NULL;
  RvpBuddy *rvpb = NULL;

  gchar *tmp = NULL;
  gchar *tmp1 = NULL;
  gchar *tmp2 = NULL;

  g_return_if_fail(hc != NULL);
  g_return_if_fail(receive != NULL);

  gc = hc->gc;
  g_return_if_fail(gc != NULL);

  session = gc->proto_data;
  g_return_if_fail(session != NULL);

  b = hc->option;
  g_return_if_fail(b != NULL);

  if (hc->inpa > 0)
    gaim_input_remove(hc->inpa);

  if (hc->fd > 0)
    close(hc->fd);

  if (strncmp(receive, "HTTP/1.1 404 Not Found\r\n", 24) == 0)
    {
      tmp = g_strdup_printf(_("buddy %s not found."),
                            gaim_buddy_get_name(b));
      gaim_notify_error(gc, NULL, _("Unable To Add"), tmp);
      gaim_blist_remove_buddy(b);
      g_free(tmp);
      return;
    }

  rvpb = b->proto_data;
  g_return_if_fail(rvpb != NULL);

  if (strncmp(receive, "HTTP/1.1 207 Multi-Status\r\n", 27) == 0)
    {
      xmlnode *node = NULL;
      xmlnode *node1 = NULL;
      xmlnode *node2 = NULL;
      RvpAwayType state = RVP_AWAY;
      gint i = 0;

      tmp = strstr(receive, "<?xml version=\"1.0\"?>");
      g_return_if_fail(tmp != NULL);

      node = xmlnode_from_str(tmp, strlen(tmp));
      g_return_if_fail(node != NULL);

      node1 = xmlnode_get_child(node,
                                "d:response/d:propstat");
      g_return_if_fail(node1 != NULL);
      node2 = xmlnode_get_child(node1,
                                "d:status");
      g_return_if_fail(node2 != NULL);

      if (strncmp(xmlnode_get_data(node2),
		  "HTTP/1.1 200 Successful", 23) != 0)
        {
          gaim_debug(GAIM_DEBUG_ERROR, "rvp", "rvp_subscribe_cb failed %s\n",
                     gaim_buddy_get_name(b));
          gaim_blist_remove_buddy(b);
	  return;
        }

      node2 = xmlnode_get_child(node1,
                                "d:prop/r:email");
      g_return_if_fail(node2 != NULL);
      tmp1 = xmlnode_get_data(node2);

      node2 = xmlnode_get_child(node1,
                                "d:prop/d:displayname");
      g_return_if_fail(node2 != NULL);
      tmp2 = xmlnode_get_data(node2);
      gaim_blist_server_alias_buddy(b, tmp2);

      node2 = xmlnode_get_child(node1,
                                "d:prop/r:state");
      g_return_if_fail(node2 != NULL);
      state = rvp_state_get_number(xmlnode_to_str(node2->child, &i));
      gaim_prpl_got_user_status(session->account, gaim_buddy_get_name(b),
                                rvp_away_get_text(state), NULL);
      if (state == RVP_IDLE)
        gaim_presence_set_idle(gaim_buddy_get_presence(b), TRUE, -1);
      else
        gaim_presence_set_idle(gaim_buddy_get_presence(b), FALSE, -1);

      xmlnode_free(node);
    }

  if ((strncmp(receive, "HTTP/1.1 207 Multi-Status\r\n", 27) == 0) ||
      (strncmp(receive, "HTTP/1.1 200 Successful\r\n", 25) == 0))
    {
      gint slt = 14400;

      gaim_debug(GAIM_DEBUG_INFO, "rvp", "rvp_subscribe_cb success %s\n",
                 gaim_buddy_get_name(b));

      if ((tmp = getHeader(receive, "Subscription-Id: ")) == NULL)
        {
          gaim_notify_error(gc, NULL, _("Subscription Id unknown"),
                            gaim_buddy_get_name(b));
          return;
        }
      rvpb->sid = atoi(tmp);
      g_free(tmp);

      if ((tmp = getHeader(receive, "Subscription-Lifetime: ")) == NULL)
        {
          slt = 14400;
        }
      else
        {
          slt = atoi(tmp);
          g_free(tmp);
        }

      if (rvpb->s_cb < 0)
        {
          rvpb->slt = slt;
          rvpb->s_cb = gaim_timeout_add((rvpb->slt - 2) * 5,
                                        (GSourceFunc) rvp_resubscribe,
                                        b);
        }
      if (rvpb->slt != slt)
        {
          rvpb->slt = slt;
          gaim_timeout_remove(rvpb->s_cb);
          rvpb->s_cb = gaim_timeout_add((rvpb->slt - 2) * 5,
                                        (GSourceFunc) rvp_resubscribe,
                                        b);
        }
    }

  return;
}

void
rvp_subscribe(RvpSession *session, GaimBuddy *buddy)
{
  GaimConnection *gc = NULL;

  RvpHttpconn *hc = g_new0(RvpHttpconn, 1);
  gchar buf[80];

  RvpBuddy *rvpb = NULL;

  g_return_if_fail(session != NULL);
  g_return_if_fail(session->connected);
  g_return_if_fail(buddy != NULL);

  g_return_if_fail(hc != NULL);

  gc = gaim_account_get_connection(session->account);
  g_return_if_fail(gc != NULL);

  gaim_debug(GAIM_DEBUG_INFO, "rvp", "rvp_subscribe %s\n",
             gaim_buddy_get_name(buddy));

  hc->gc = gc;
  hc->connection_step = 1;

  if (gc->inpa > 0)
    gaim_input_remove(gc->inpa);

  if (buddy->proto_data == NULL)
    {
      gchar **mail = NULL;

      buddy->proto_data = g_new0(RvpBuddy, 1);
      rvpb = buddy->proto_data;
      g_return_if_fail(rvpb != NULL);

      mail = g_strsplit(gaim_buddy_get_name(buddy), "@", -1);
      g_return_if_fail(mail != NULL);

      rvpb->host = mail[1];
      rvpb->user = mail[0];
      g_free(mail);

      rvpb->sid = 0;
      rvpb->slt = 14400;
      rvpb->session = session;
      rvpb->sessionid = NULL;
      rvpb->s_cb = -1;
      rvpb->xfer = NULL;
    }
  else
    {
      rvpb = buddy->proto_data;
      g_return_if_fail(rvpb != NULL);
    }

  hc->method = "SUBSCRIBE";
  hc->uri = g_strdup_printf(
                            "/instmsg/aliases/%s",
                            rvpb->user);
  hc->callback = FALSE;
  hc->host = g_strdup(rvpb->host);
  hc->option = (gpointer *)buddy;
  hc->body = g_strdup("");

  hc->header = g_strdup_printf(
			       "%s %s HTTP/1.1\r\n"
			       "Subscription-Lifetime: %d\r\n"
			       "Notification-Type: update/propchange\r\n"
                               "Call-Back: http://%s/instmsg/aliases/%s\r\n"
			       "Host: %s\r\n"
			       "RVP-Notifications-Version: 0.2\r\n"
			       "RVP-From-Principal: http://%s/instmsg/aliases/%s\r\n"
                               "Content-Length: 0\r\n",
                               hc->method,
			       hc->uri,
                               rvpb->slt,
                               session->host,
                               session->user,
			       hc->host,
			       session->host,
			       session->user);
  hc->analyze = rvp_subscribe_cb;

  if (gaim_proxy_connect(session->account, hc->host, RVP_PORT,
                         rvp_http_send, hc) != 0)
    {
      g_snprintf(buf, sizeof buf, _("buddy %s not found."),
                 gaim_buddy_get_name(buddy));
      gaim_notify_error(gc, NULL, _("Unable To Add"), buf);
      gaim_blist_remove_buddy(buddy);
    }

  return;
}

gboolean
rvp_resubscribe(GaimBuddy *buddy)
{
  RvpBuddy *rvpb = NULL;

  g_return_val_if_fail(buddy != NULL, FALSE);

  rvpb = buddy->proto_data;
  g_return_val_if_fail(rvpb != NULL, FALSE);

  gaim_debug(GAIM_DEBUG_INFO, "rvp", "rvp_resubscribe %s\n",
             gaim_buddy_get_name(buddy));
  rvp_subscribe(rvpb->session, buddy);
  return TRUE;
}


void
rvp_unsubscribe_cb(RvpHttpconn *hc, const gchar *receive)
{
  g_return_if_fail(hc != NULL);
  g_return_if_fail(receive != NULL);

  if (hc->inpa > 0)
    gaim_input_remove(hc->inpa);

  if (hc->fd > 0)
    close(hc->fd);

  if (strncmp(receive, "HTTP/1.1 200 Successful\r\n", 25) == 0)
    gaim_debug(GAIM_DEBUG_INFO, "rvp", "rvp_unsubscribe_cb success\n");
  else
    gaim_debug(GAIM_DEBUG_ERROR, "rvp", "rvp_unsubscribe_cb failed\n");

  return;
}

void
rvp_unsubscribe(RvpSession *session, GaimBuddy *buddy)
{
  GaimConnection *gc = NULL;

  RvpHttpconn *hc = g_new0(RvpHttpconn, 1);
  gchar buf[80];

  RvpBuddy *rvpb = NULL;

  g_return_if_fail(session != NULL);
  g_return_if_fail(session->connected);
  g_return_if_fail(buddy != NULL);

  g_return_if_fail(hc != NULL);

  gc = gaim_account_get_connection(session->account);
  g_return_if_fail(gc != NULL);

  rvpb = buddy->proto_data;
  g_return_if_fail(rvpb != NULL);

  gaim_debug(GAIM_DEBUG_INFO, "rvp", "rvp_unsubscribe %s\n",
             gaim_buddy_get_name(buddy));

  hc->gc = gc;
  hc->connection_step = 1;

  if (gc->inpa > 0)
    gaim_input_remove(gc->inpa);

  gaim_timeout_remove(rvpb->s_cb);

  hc->method = "UNSUBSCRIBE";
  hc->uri = g_strdup_printf(
                            "/instmsg/aliases/%s",
                            rvpb->user);
  hc->callback = FALSE;
  hc->option = NULL;
  hc->host = g_strdup(rvpb->host);
  hc->body = g_strdup("");

  hc->header = g_strdup_printf(
                               "%s %s HTTP/1.1\r\n"
                               "Host: %s\r\n"
                               "RVP-Notifications-Version: 0.2\r\n"
                               "RVP-From-Principal: http://%s/instmsg/aliases/%s\r\n"
                               "Subscription-Id: %d\r\n"
                               "Content-Length: 0\r\n",
                               hc->method,
                               hc->uri,
                               hc->host,
                               session->host,
                               session->user,
                               rvpb->sid);
  hc->analyze = rvp_unsubscribe_cb;

  if (gaim_proxy_connect(session->account, hc->host, RVP_PORT,
                         rvp_http_send, hc) != 0)
    {
      g_snprintf(buf, sizeof buf, _("buddy %s not found."),
                 gaim_buddy_get_name(buddy));
      gaim_notify_error(gc, NULL, _("Unable To Add"), buf);
      gaim_blist_remove_buddy(buddy);
    }

  return;
}
